import numba as nb
import numpy as np

MIN_FLOAT64 = np.finfo(np.float64).min


@nb.njit(cache=False)
def _make_dtw_matrix(
    score_matrix: np.ndarray,
    gap_open_penalty: float = 0.0,
    gap_extend_penalty: float = 0.0,
):
    """
    Make cost matrix using dynamic time warping

    Parameters
    ----------
    score_matrix
        matrix of scores between corresponding vectors of the two vector sets; shape = (n, m)
    gap_open_penalty
        penalty for opening a (series of) gap(s)
    gap_extend_penalty
        penalty for extending an existing series of gaps

    Returns
    -------
    accumulated cost matrix; shape = (n, m)
    """
    gap_open_penalty *= -1
    gap_extend_penalty *= -1
    n, m = score_matrix.shape
    matrix = np.zeros((n + 1, m + 1, 3), dtype=np.float64)
    matrix[:, 0, :] = MIN_FLOAT64
    matrix[0, :, :] = MIN_FLOAT64
    matrix[0, 0] = 0
    backtrack = np.zeros((n + 1, m + 1, 3), dtype=np.int64)
    for i in range(1, n + 1):
        matrix[i, 0, 0] = 0
        matrix[i, 0, 1] = 0
        matrix[i, 0, 2] = MIN_FLOAT64 - gap_open_penalty
        backtrack[i, 0] = 0

    for j in range(1, m + 1):
        matrix[0, j, 0] = MIN_FLOAT64 - gap_open_penalty
        matrix[0, j, 1] = 0
        matrix[0, j, 2] = 0
        backtrack[0, j] = 1

    for i in range(1, n + 1):
        for j in range(1, m + 1):
            scores_lower = np.array(
                [
                    matrix[i - 1, j, 0] + gap_extend_penalty,
                    matrix[i - 1, j, 1] + gap_open_penalty,
                ]
            )
            index_lower = np.argmax(scores_lower)
            matrix[i, j, 0] = scores_lower[index_lower]
            backtrack[i, j, 0] = index_lower

            scores_upper = np.array(
                [
                    matrix[i, j - 1, 1] + gap_open_penalty,
                    matrix[i, j - 1, 2] + gap_extend_penalty,
                ]
            )
            index_upper = np.argmax(scores_upper)
            matrix[i, j, 2] = scores_upper[index_upper]
            backtrack[i, j, 2] = index_upper + 1

            scores = np.array(
                [
                    matrix[i, j, 0],
                    matrix[i - 1, j - 1, 1] + score_matrix[i - 1, j - 1],
                    matrix[i, j, 2],
                ]
            )
            index = np.argmax(scores)
            matrix[i, j, 1] = scores[index]
            backtrack[i, j, 1] = index
    return matrix, backtrack


@nb.njit(cache=False)
def _get_dtw_alignment(start_direction, backtrack: np.ndarray, n1, m1):
    """
    Finds optimal warping path from a backtrack matrix

    Parameters
    ----------
    start_direction
    backtrack
    n1
        length of first sequence
    m1
        length of second sequence

    Returns
    -------
    aligned_indices_1, aligned_indices_2
    """
    indices_1 = np.zeros(n1 + m1 + 1, dtype=np.int64)
    indices_2 = np.zeros(n1 + m1 + 1, dtype=np.int64)
    index = 0
    n, m = n1, m1
    direction = start_direction
    while not (n == 0 and m == 0):
        if m == 0:
            n -= 1
            indices_1[index] = n
            indices_2[index] = -1
            index += 1
        elif n == 0:
            m -= 1
            indices_1[index] = -1
            indices_2[index] = m
            index += 1
        else:
            if direction == 0:
                direction = backtrack[n, m, 0]
                n -= 1
                indices_1[index] = n
                indices_2[index] = -1
                index += 1
            elif direction == 1:
                direction = backtrack[n, m, 1]
                if direction == 1:
                    n -= 1
                    m -= 1
                    indices_1[index] = n
                    indices_2[index] = m
                    index += 1
            elif direction == 2:
                direction = backtrack[n, m, 2]
                m -= 1
                indices_1[index] = -1
                indices_2[index] = m
                index += 1
    return indices_1[:index][::-1], indices_2[:index][::-1]


@nb.njit
def dtw_align(
    score_matrix: np.ndarray,
    gap_open_penalty: float = 0.0,
    gap_extend_penalty: float = 0.0,
):
    """
    Align two objects using dynamic time warping

    Parameters
    ----------
    score_matrix
        (n x m) matrix of scores between all points of both objects
    gap_open_penalty
        penalty for opening a (series of) gap(s)
    gap_extend_penalty
        penalty for extending an existing series of gaps
    Returns
    -------
    aligned_indices_1, aligned_indices_2
    """
    matrix, backtrack = _make_dtw_matrix(
        score_matrix, gap_open_penalty, gap_extend_penalty
    )
    n = score_matrix.shape[0]
    m = score_matrix.shape[1]
    scores = np.array([matrix[n, m, 0], matrix[n, m, 1], matrix[n, m, 2]])
    index = np.argmax(scores)
    aln_1, aln_2 = _get_dtw_alignment(index, backtrack, n, m)
    return aln_1, aln_2, scores[index]



import pickle
import typing
from dataclasses import dataclass
from pathlib import Path

import numba as nb
import numpy as np
import prody as pd
import typer
from geometricus import Structure, MomentInvariants, SplitType, GeometricusEmbedding
from scipy.spatial.distance import pdist, squareform
from copy import deepcopy
import arviz as az

from caretta import (
    dynamic_time_warping as dtw,
    neighbor_joining as nj,
    score_functions,
    superposition_functions,
    feature_extraction,
    helper,
)


def alignment_to_numpy(alignment):
    aln_np = {}
    for n in alignment:
        aln_seq = []
        index = 0
        for a in alignment[n]:
            if a == "-":
                aln_seq.append(-1)
            else:
                aln_seq.append(index)
                index += 1
        aln_np[n] = np.array(aln_seq)
    return aln_np


@nb.njit
def tm_score(coords_1, coords_2, l1, l2):
    d1 = 1.24 * (l1 - 15) ** 1 / 3 - 1.8
    d2 = 1.24 * (l2 - 15) ** 1 / 3 - 1.8
    sum_1 = 0
    sum_2 = 0
    for i in range(coords_1.shape[0]):
        sum_1 += 1 / (1 + (np.sqrt(np.sum((coords_1[i] - coords_2[i]) ** 2)) / d1))
        sum_2 += 1 / (1 + (np.sqrt(np.sum((coords_1[i] - coords_2[i]) ** 2)) / d2))
    t1 = (1 / l1) * sum_1
    t2 = (1 / l2) * sum_2
    return max(t1, t2)


@nb.njit
def get_common_coordinates(
    coords_1: np.ndarray, coords_2: np.ndarray, aln_1: np.ndarray, aln_2: np.ndarray
) -> typing.Tuple[np.ndarray, np.ndarray]:
    """
    Return coordinate positions aligned in both coords_1 and coords_2
    """
    assert aln_1.shape == aln_2.shape
    pos_1, pos_2 = helper.get_common_positions(aln_1, aln_2)
    return coords_1[pos_1], coords_2[pos_2]


@nb.njit
def get_mean_coords(
    aln_1, coords_1: np.ndarray, aln_2, coords_2: np.ndarray
) -> np.ndarray:
    """
    Mean of two coordinate sets (of the same shape)
    """
    mean_coords = np.zeros((aln_1.shape[0], coords_1.shape[1]))
    for i, (x, y) in enumerate(zip(aln_1, aln_2)):
        if x == -1:
            mean_coords[i] = coords_2[y]
        elif y == -1:
            mean_coords[i] = coords_1[x]
        else:
            mean_coords[i] = np.array(
                [
                    np.nanmean(np.array([coords_1[x, d], coords_2[y, d]]))
                    for d in range(coords_1.shape[1])
                ]
            )
    return mean_coords


def get_mean_weights(
    weights_1: np.ndarray, weights_2: np.ndarray, aln_1: np.ndarray, aln_2: np.ndarray
) -> np.ndarray:
    mean_weights = np.zeros(aln_1.shape[0])
    for i, (x, y) in enumerate(zip(aln_1, aln_2)):
        if not x == -1:
            mean_weights[i] += weights_1[x]
        if not y == -1:
            mean_weights[i] += weights_2[y]
    return mean_weights


@nb.njit
def get_pairwise_alignment(
    coords_1,
    coords_2,
    gamma,
    gap_open_penalty: float,
    gap_extend_penalty: float,
    weights_1: np.ndarray,
    weights_2: np.ndarray,
    n_iter=3,
):
    score_matrix = score_functions.make_score_matrix(
        np.hstack((coords_1, weights_1)),
        np.hstack((coords_2, weights_2)),
        score_functions.get_caretta_score,
        gamma,
    )
    dtw_aln_array_1, dtw_aln_array_2, dtw_score = dtw.dtw_align(
        score_matrix, gap_open_penalty, gap_extend_penalty
    )
    for i in range(n_iter):
        pos_1, pos_2 = helper.get_common_positions(dtw_aln_array_1, dtw_aln_array_2)
        common_coords_1, common_coords_2 = coords_1[pos_1], coords_2[pos_2]
        (
            c1,
            c2,
            common_coords_2,
        ) = superposition_functions.paired_svd_superpose_with_subset(
            coords_1, coords_2, common_coords_1, common_coords_2
        )
        score_matrix = score_functions.make_score_matrix(
            np.hstack((c1, weights_1)),
            np.hstack((c2, weights_2)),
            score_functions.get_caretta_score,
            gamma,
        )
        aln_1, aln_2, score = dtw.dtw_align(
            score_matrix, gap_open_penalty, gap_extend_penalty
        )
        if score > dtw_score:
            coords_1 = c1
            coords_2 = c2
            dtw_score = score
            dtw_aln_array_1 = aln_1
            dtw_aln_array_2 = aln_2
        else:
            break
    return dtw_aln_array_1, dtw_aln_array_2, dtw_score, coords_1, coords_2


@dataclass
class OutputFiles:
    fasta_file: Path = Path("./result.fasta")
    pdb_folder: Path = Path("./result_pdb/")
    cleaned_pdb_folder: Path = Path("./cleaned_pdb")
    feature_file: Path = Path("./result_features.pkl")
    class_file: Path = Path("./result_class.pkl")


DEFAULT_SUPERPOSITION_PARAMETERS = {
    # must-have
    "gap_open_penalty": 0.0,
    "gap_extend_penalty": 0.0,
    "gamma": 0.03,
    # changes per superposition_function
    "split_type": "KMER",
    "split_size": 20,
    "scale": True,
    "gamma_moment": 0.6,
    "n_iter": 3,
}


@dataclass
class StructureMultiple:
    """
    Class for multiple structure alignment

    Constructor Arguments
    ---------------------
    structures
        list of protein_utility.Structure objects
    superposition_parameters
        dictionary of parameters to pass to the superposition function
    superposition_function
        a function that takes two coordinate sets as input and superposes them
        returns a score, superposed_coords_1, superposed_coords_2
    score_function
        a function that takes two paired coordinate sets (same shape) and returns a score
    consensus_weight
        weights the effect of well-aligned columns on the progressive alignment
    final_structures
        makes the progressive alignment tree of structures, last one is the consensus structure of the multiple alignment
    tree
        neighbor joining tree of indices controlling alignment - indices beyond len(structures) refer to intermediate nodes
    branch_lengths
    alignment
        indices of aligning residues from each structure, gaps are -1s
    """

    structures: typing.List[Structure]
    sequences: typing.Dict[str, str]
    superposition_parameters: typing.Dict[str, typing.Any]
    superposition_function: typing.Callable[
        [
            np.ndarray,
            np.ndarray,
            dict,
            typing.Callable[[np.ndarray, np.ndarray], float],
        ],
        typing.Tuple[float, np.ndarray, np.ndarray],
    ] = lambda x, y: (0, x, y)
    mean_function: typing.Callable[
        [np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray
    ] = get_mean_coords
    consensus_weight: float = 1.0
    pairwise_distance_matrix: typing.Union[None, np.ndarray] = None
    reference_structure_index: typing.Union[None, int] = None
    final_structures: typing.Union[None, typing.List[Structure]] = None
    final_consensus_weights: typing.Union[None, typing.List[np.ndarray]] = None
    tree: typing.Union[None, np.ndarray] = None
    branch_lengths: typing.Union[None, np.ndarray] = None
    alignment: typing.Union[dict, None] = None
    output_folder: Path = Path("./caretta_results")
    features: typing.Union[dict, None] = None

    @staticmethod
    def align_from_pdb_files(
        input_pdb: typing.Union[typing.List[str], Path, str],
        gap_open_penalty: float = 1.0,
        gap_extend_penalty: float = 0.01,
        consensus_weight: float = 1.0,
        full: bool = False,
        output_folder: typing.Union[str, Path] = Path("../caretta_results"),
        num_threads: int = 20,
        write_fasta: bool = False,
        write_pdb: bool = False,
        write_features: bool = False,
        only_dssp: bool = True,
        write_class: bool = False,
        write_matrix: bool = False,
        verbose: bool = True,
    ):
        """
        Caretta aligns protein structures and can output a sequence alignment, superposed PDB files,
        a set of aligned feature matrices and a class with intermediate structures made during progressive alignment.

        Parameters
        ----------
        input_pdb
            Can be \n
            A list of PDB files,
            A list of PDB IDs,
            A folder with input protein files,
            A file which lists PDB filenames on each line,
            A file which lists PDB IDs on each line,
        gap_open_penalty
            default 1
        gap_extend_penalty
            default 0.01
        consensus_weight
            default 1
        full
            True =>  Uses all-vs-all pairwise Caretta alignment to make the distance matrix (much slower)
        output_folder
            default "caretta_results"
        num_threads
            Number of threads to use for feature extraction
        write_fasta
            True => writes alignment as fasta file (default True)
            writes to output_folder / result.fasta
        write_pdb
            True => writes all protein PDB files superposed by alignment (default True)
             writes to output_folder / superposed_pdb
        write_features
            True => extracts and writes aligned features as a dictionary of numpy arrays into a pickle file (default True)
            writes to output_folder / result_features.pkl
        only_dssp
            True => only DSSP features extracted
            False => all features
        write_class
            True => writes StructureMultiple class with intermediate structures and tree to pickle file (default True)
            writes to output_folder / result_class.pkl
        write_matrix
            True => writes distance matrix to text file (default False)
            writes to output_folder / matrix.mat
        verbose
            controls verbosity

        Returns
        -------
        StructureMultiple class
        """
        msa_class = StructureMultiple.from_pdb_files(
            input_pdb,
            superposition_parameters=DEFAULT_SUPERPOSITION_PARAMETERS,
            superposition_function=superposition_functions.moment_svd_superpose_function,
            consensus_weight=consensus_weight,
            output_folder=output_folder,
            verbose=verbose,
        )
        if len(msa_class.structures) > 2:
            if full:
                msa_class.make_pairwise_dtw_matrix(
                    gap_open_penalty, gap_extend_penalty, verbose=verbose
                )
            else:
                msa_class.make_pairwise_shape_matrix(
                    verbose=verbose
                )
            msa_class.align(gap_open_penalty, gap_extend_penalty, verbose=verbose)
        else:
            msa_class.align(
                gap_open_penalty=gap_open_penalty,
                gap_extend_penalty=gap_extend_penalty,
                verbose=verbose,
            )

        msa_class.write_files(
            write_fasta=write_fasta,
            write_pdb=write_pdb,
            write_features=write_features,
            write_class=write_class,
            write_matrix=write_matrix,
            num_threads=num_threads,
            only_dssp=only_dssp,
            verbose=verbose,
        )
        return msa_class

    @classmethod
    def from_pdb_files(
        cls,
        input_pdb,
        superposition_parameters=DEFAULT_SUPERPOSITION_PARAMETERS,
        superposition_function=superposition_functions.moment_svd_superpose_function,
        consensus_weight=1.0,
        output_folder=Path("./caretta_results"),
        verbose: bool = False,
    ):
        """
        Makes a StructureMultiple object from a list of pdb files/names or a folder of pdb files

        Parameters
        ----------
        input_pdb
            list of pdb files/names or a folder containing pdb files
        superposition_parameters
            parameters to give to the superposition function
        superposition_function
            a function that takes two coordinate sets as input and superposes them
            returns a score, superposed_coords_1, superposed_coords_2
        consensus_weight
            weights the effect of well-aligned columns on the progressive alignment
        output_folder
        verbose

        Returns
        -------
        StructureMultiple object (unaligned)
        """
        output_folder = Path(output_folder)
        if not output_folder.exists():
            output_folder.mkdir()

        cleaned_pdb_folder = output_folder / "cleaned_pdb"
        if not cleaned_pdb_folder.exists():
            cleaned_pdb_folder.mkdir()
        pdb_files = helper.parse_pdb_files_and_clean(input_pdb, cleaned_pdb_folder)
        if verbose:
            typer.echo(f"Found {len(pdb_files)} PDB files")

        structures = []
        sequences = {}
        for pdb_file in pdb_files:
            pdb_name = Path(pdb_file).stem
            protein = pd.parsePDB(str(pdb_file)).select("protein and calpha")
            coordinates = protein.getCoords()
            structures.append(Structure(pdb_name, coordinates.shape[0], coordinates))
            sequences[pdb_name] = protein.getSequence()
        msa_class = StructureMultiple(
            structures,
            sequences,
            superposition_parameters,
            superposition_function,
            consensus_weight=consensus_weight,
            output_folder=output_folder,
        )
        return msa_class

    @classmethod
    def from_coordinates(
        cls,
        names: typing.List[str],
        coordinates_list: typing.List[np.ndarray],
        sequences: typing.List[str],
        superposition_parameters: dict,
        superposition_function=superposition_functions.moment_multiple_svd_superpose_function,
        consensus_weight=1.0,
        output_folder=Path("./caretta_results"),
    ):
        """
        Makes a StructureMultiple object from a list of coordinates

        Parameters
        ----------
        names
        coordinates_list
        sequences
        superposition_parameters
            parameters to give to the superposition function
        superposition_function
            a function that takes two coordinate sets as input and superposes them
            returns a score, superposed_coords_1, superposed_coords_2
        consensus_weight
            weights the effect of well-aligned columns on the progressive alignment
        output_folder

        Returns
        -------
        StructureMultiple object (unaligned)
        """
        output_folder = Path(output_folder)
        if not output_folder.exists():
            output_folder.mkdir()
        sequences = {n: s for n, s in zip(names, sequences)}
        structures = []
        for name, coordinates in zip(names, coordinates_list):
            structures.append(Structure(name, coordinates.shape[0], coordinates))
        msa_class = StructureMultiple(
            structures,
            sequences,
            superposition_parameters,
            superposition_function,
            consensus_weight=consensus_weight,
            output_folder=output_folder,
        )
        return msa_class

    def get_pairwise_alignment(
        self,
        coords_1,
        coords_2,
        gap_open_penalty: float,
        gap_extend_penalty: float,
        weight=False,
        weights_1=None,
        weights_2=None,
        n_iter=3,
        verbose: bool = False,
    ):
        """
        Aligns coords_1 to coords_2 by first superposing and then running dtw on the score matrix of the superposed coordinate sets
        Parameters
        ----------
        coords_1
        coords_2
        gap_open_penalty
        gap_extend_penalty
        weight
        weights_1
        weights_2
        n_iter
        verbose

        Returns
        -------
        alignment_1, alignment_2, score, superposed_coords_1, superposed_coords_2
        """
        _, coords_1, coords_2 = self.superposition_function(
            coords_1, coords_2, self.superposition_parameters
        )

        if weight:
            assert weights_1 is not None
            assert weights_2 is not None
            weights_1 = weights_1.reshape(-1, 1)
            weights_2 = weights_2.reshape(-1, 1)
        else:
            weights_1 = np.zeros((coords_1.shape[0], 1))
            weights_2 = np.zeros((coords_2.shape[0], 1))
        return get_pairwise_alignment(
            coords_1,
            coords_2,
            self.superposition_parameters["gamma"],
            gap_open_penalty,
            gap_extend_penalty,
            weights_1,
            weights_2,
            n_iter,
        )

    def make_pairwise_shape_matrix(
        self,
        resolution: typing.Union[float, np.ndarray] = 2.0,
        parameters: dict = None,
        metric="braycurtis",
        verbose: bool = False,
    ):
        """
        Makes an all vs. all matrix of distance scores between all the structures.

        Parameters
        ----------
        resolution
        parameters
            to use for making invariants
            needs to have
            num_split_types
            split_type_i
            split_size_i
        metric
            distance metric (accepts any metric supported by scipy.spatial.distance
        verbose
        Returns
        -------
        [n x n] distance matrix
        """
        if verbose:
            typer.echo("Calculating pairwise distances...")
        if parameters is None:
            parameters = dict(num_split_types=1, split_type_0="KMER", split_size_0=20)
        embedders = []
        for i in range(parameters["num_split_types"]):
            invariants = (
                MomentInvariants.from_coordinates(
                    s.name,
                    s.coordinates,
                    None,
                    split_size=parameters[f"split_size_{i}"],
                    split_type=SplitType[parameters[f"split_type_{i}"]],
                )
                for s in self.structures
            )
            embedders.append(
                GeometricusEmbedding.from_invariants(
                    invariants,
                    resolution=resolution,
                    protein_keys=[s.name for s in self.structures],
                )
            )
        distance_matrix = squareform(
            pdist(
                np.hstack([embedder.embedding for embedder in embedders]),
                metric=metric,
            )
        )
        self.reference_structure_index = np.argmin(
            np.median(distance_matrix, axis=0)
        )
        self.pairwise_distance_matrix = distance_matrix

    def make_pairwise_dtw_matrix(
        self,
        gap_open_penalty: float,
        gap_extend_penalty: float,
        invert=True,
        verbose: bool = False,
    ):
        """
        Makes an all vs. all matrix of distance (or similarity) scores between all the structures using pairwise alignment.

        Parameters
        ----------
        gap_open_penalty
        gap_extend_penalty
        invert
            if True returns distance matrix
            if False returns similarity matrix
        verbose
        Returns
        -------
        [n x n] matrix
        """
        if verbose:
            typer.echo("Calculating pairwise distances...")
        pairwise_matrix = np.zeros((len(self.structures), len(self.structures)))
        for i in range(pairwise_matrix.shape[0] - 1):
            for j in range(i + 1, pairwise_matrix.shape[1]):
                coords_1, coords_2 = (
                    self.structures[i].coordinates,
                    self.structures[j].coordinates,
                )
                (
                    dtw_aln_1,
                    dtw_aln_2,
                    score,
                    coords_1,
                    coords_2,
                ) = self.get_pairwise_alignment(
                    coords_1,
                    coords_2,
                    gap_open_penalty=gap_open_penalty,
                    gap_extend_penalty=gap_extend_penalty,
                    weight=False,
                )
                common_coords_1, common_coords_2 = get_common_coordinates(
                    coords_1, coords_2, dtw_aln_1, dtw_aln_2
                )
                pairwise_matrix[i, j] = score_functions.get_total_score(
                    common_coords_1,
                    common_coords_2,
                    score_functions.get_caretta_score,
                    self.superposition_parameters["gamma"],
                    True,
                )
                if invert:
                    pairwise_matrix[i, j] *= -1
        pairwise_matrix += pairwise_matrix.T
        self.reference_structure_index = np.argmin(
            np.median(pairwise_matrix, axis=0)
        )
        self.pairwise_distance_matrix = pairwise_matrix

    def align(
        self,
        gap_open_penalty,
        gap_extend_penalty,
        return_sequence: bool = True,
        verbose: bool = False,
    ) -> dict:
        """
        Makes a multiple structure alignment

        Parameters
        ----------
        pw_matrix
            pairwise similarity matrix to base the neighbor joining tree on
        gap_open_penalty
        gap_extend_penalty
        return_sequence
            if True returns sequence alignment
            else indices of aligning residues with gaps as -1s
        verbose
        Returns
        -------
        alignment = {name: indices of aligning residues with gaps as -1s}
        """
        if len(self.structures) == 2:
            coords_1, coords_2 = (
                self.structures[0].coordinates,
                self.structures[1].coordinates,
            )
            dtw_1, dtw_2, _, _, _ = self.get_pairwise_alignment(
                coords_1,
                coords_2,
                gap_open_penalty=gap_open_penalty,
                gap_extend_penalty=gap_extend_penalty,
                weight=False,
                n_iter=self.superposition_parameters["n_iter"],
                verbose=verbose,
            )
            self.alignment = {
                self.structures[0].name: dtw_1,
                self.structures[1].name: dtw_2,
            }
            if return_sequence:
                return self.make_sequence_alignment()
            else:
                return self.alignment
        assert self.pairwise_distance_matrix is not None
        assert self.pairwise_distance_matrix.shape[0] == len(self.structures)
        if verbose:
            typer.echo("Constructing neighbor joining tree...")
        tree, branch_lengths = nj.neighbor_joining(self.pairwise_distance_matrix)
        self.tree = tree
        self.branch_lengths = branch_lengths
        self.final_structures = [s for s in self.structures]
        self.final_consensus_weights = [
            np.full(
                (self.structures[i].coordinates.shape[0], 1),
                self.consensus_weight,
                dtype=np.float64,
            )
            for i in range(len(self.structures))
        ]
        msa_alignments = {
            s.name: {s.name: np.arange(s.length)} for s in self.final_structures
        }

        def make_intermediate_node(n1, n2, n_int):
            name_1, name_2 = (
                self.final_structures[n1].name,
                self.final_structures[n2].name,
            )

            name_int = f"int-{n_int}"
            n1_coords = self.final_structures[n1].coordinates
            n1_weights = self.final_consensus_weights[n1]
            n1_weights *= len(msa_alignments[name_2])
            n1_weights /= 2 * (
                len(msa_alignments[name_2]) + len(msa_alignments[name_1])
            )
            n2_coords = self.final_structures[n2].coordinates
            n2_weights = self.final_consensus_weights[n2]
            n2_weights *= len(msa_alignments[name_1])
            n2_weights /= 2 * (
                len(msa_alignments[name_2]) + len(msa_alignments[name_1])
            )
            (
                dtw_aln_1,
                dtw_aln_2,
                score,
                n1_coords,
                n2_coords,
            ) = self.get_pairwise_alignment(
                n1_coords,
                n2_coords,
                gap_open_penalty=gap_open_penalty,
                gap_extend_penalty=gap_extend_penalty,
                weight=True,
                weights_1=n1_weights,
                weights_2=n2_weights,
                n_iter=self.superposition_parameters["n_iter"],
            )
            n1_weights *= (
                2
                * (len(msa_alignments[name_2]) + len(msa_alignments[name_1]))
                / len(msa_alignments[name_2])
            )
            n2_weights *= (
                2
                * (len(msa_alignments[name_2]) + len(msa_alignments[name_1]))
                / len(msa_alignments[name_1])
            )
            msa_alignments[name_1] = {
                name: np.array([sequence[i] if i != -1 else -1 for i in dtw_aln_1])
                for name, sequence in msa_alignments[name_1].items()
            }
            msa_alignments[name_2] = {
                name: np.array([sequence[i] if i != -1 else -1 for i in dtw_aln_2])
                for name, sequence in msa_alignments[name_2].items()
            }
            msa_alignments[name_int] = {
                **msa_alignments[name_1],
                **msa_alignments[name_2],
            }

            mean_coords = self.mean_function(dtw_aln_1, n1_coords, dtw_aln_2, n2_coords)
            mean_weights = get_mean_weights(
                n1_weights, n2_weights, dtw_aln_1, dtw_aln_2
            )
            self.final_structures.append(
                Structure(name_int, mean_coords.shape[0], mean_coords)
            )
            self.final_consensus_weights.append(mean_weights)

        if verbose:
            with typer.progressbar(
                range(0, self.tree.shape[0] - 1, 2), label="Aligning"
            ) as progress:
                for x in progress:
                    node_1, node_2, node_int = (
                        self.tree[x, 0],
                        self.tree[x + 1, 0],
                        self.tree[x, 1],
                    )
                    assert self.tree[x + 1, 1] == node_int
                    make_intermediate_node(node_1, node_2, node_int)
        else:
            for x in range(0, self.tree.shape[0] - 1, 2):
                node_1, node_2, node_int = (
                    self.tree[x, 0],
                    self.tree[x + 1, 0],
                    self.tree[x, 1],
                )
                assert self.tree[x + 1, 1] == node_int
                make_intermediate_node(node_1, node_2, node_int)

        node_1, node_2 = self.tree[-1, 0], self.tree[-1, 1]
        make_intermediate_node(node_1, node_2, "final")
        alignment = {
            **msa_alignments[self.final_structures[node_1].name],
            **msa_alignments[self.final_structures[node_2].name],
        }
        self.alignment = alignment
        if return_sequence:
            return self.make_sequence_alignment()
        else:
            return alignment

    def make_sequence_alignment(self, alignment=None):
        sequence_alignment = {}
        if alignment is None:
            alignment = self.alignment
        for s in self.structures:
            sequence_alignment[s.name] = "".join(
                self.sequences[s.name][i] if i != -1 else "-" for i in alignment[s.name]
            )
        return sequence_alignment

    def write_files(
        self,
        write_fasta,
        write_pdb,
        write_features,
        write_class,
        write_matrix,
        only_dssp=True,
        num_threads=4,
        verbose: bool = False,
    ):
        if verbose and any(
            (write_fasta, write_pdb, write_pdb, write_class, write_matrix)
        ):
            typer.echo("Writing files...")
        if write_fasta:
            fasta_file = self.output_folder / "result.fasta"
            self.write_alignment(fasta_file)
            if verbose:
                typer.echo(
                    f"FASTA file: {typer.style(str(fasta_file), fg=typer.colors.GREEN)}",
                )
        if write_pdb:
            pdb_folder = self.output_folder / "superposed_pdbs"
            if not pdb_folder.exists():
                pdb_folder.mkdir()
            self.write_superposed_pdbs(pdb_folder, verbose=verbose)
            if verbose:
                typer.echo(
                    f"Superposed PDB files: {typer.style(str(pdb_folder), fg=typer.colors.GREEN)}"
                )
        if write_features:
            dssp_dir = self.output_folder / ".caretta_tmp"
            if not dssp_dir.exists():
                dssp_dir.mkdir()
            feature_file = self.output_folder / "result_features.pkl"
            self.features = self.get_aligned_features(
                str(dssp_dir), only_dssp=only_dssp, num_threads=num_threads
            )
            with open(feature_file, "wb") as f:
                pickle.dump(self.features, f)
            if verbose:
                typer.echo(
                    f"Aligned features: {typer.style(str(feature_file), fg=typer.colors.GREEN)}"
                )
        if write_class:
            class_file = self.output_folder / "result_class.pkl"
            with open(class_file, "wb") as f:
                pickle.dump(self, f)
            if verbose:
                typer.echo(
                    f"Class file: {typer.style(str(class_file), fg=typer.colors.GREEN)}"
                )
        if write_matrix:
            matrix_file = self.output_folder / "matrix.mat"
            helper.write_distance_matrix(
                [s.name for s in self.structures],
                self.pairwise_distance_matrix,
                matrix_file,
            )
            if verbose:
                typer.echo(
                    f"Distance matrix file: {typer.style(str(matrix_file), fg=typer.colors.GREEN)}"
                )

    def write_alignment(self, filename, alignments: dict = None):
        """
        Writes alignment to a fasta file
        """
        if alignments is None:
            alignments = self.alignment
        with open(filename, "w") as f:
            for key in alignments:
                sequence = "".join(
                    self.sequences[key][n] if n != -1 else "-" for n in alignments[key]
                )
                f.write(f">{key}\n{sequence}\n")

    def write_superposed_pdbs(
        self, output_pdb_folder, alignments: dict = None, verbose: bool = False
    ):
        """
        Superposes PDBs according to alignment and writes transformed PDBs to files
        (View with Pymol)

        Parameters
        ----------
        output_pdb_folder
        alignments
        verbose
        """
        if alignments is None:
            alignments = self.alignment
        output_pdb_folder = Path(output_pdb_folder)
        if not output_pdb_folder.exists():
            output_pdb_folder.mkdir()
        reference_name = self.structures[self.reference_structure_index].name
        core_indices = np.array(
            [
                i
                for i in range(len(alignments[reference_name]))
                if -1 not in [alignments[n][i] for n in alignments]
            ]
        )
        if verbose:
            typer.echo(
                f"{len(core_indices)} core positions in alignment of length {len(alignments[reference_name])}"
            )
        if len(core_indices) < len(alignments[reference_name]) // 2:
            if verbose:
                typer.echo(
                    typer.style(
                        "Core indices are < half of alignment length, superposing using reference structure "
                        "instead",
                        fg=typer.colors.RED,
                    )
                )
                typer.echo(
                    typer.style(
                        "Please inspect the distance matrix to split divergent protein groups",
                        fg=typer.colors.RED,
                    )
                )
            self.write_superposed_pdbs_reference(output_pdb_folder, alignments)
        else:
            self.write_superposed_pdbs_core(output_pdb_folder, alignments)

    def write_superposed_pdbs_core(self, output_pdb_folder, alignments):
        """
        Superposes PDBs according to core indices in alignment and writes transformed PDBs to files
        (View with Pymol)

        Parameters
        ----------
        alignments
        output_pdb_folder
        """
        reference_name = self.structures[self.reference_structure_index].name
        core_indices = np.array(
            [
                i
                for i in range(len(alignments[reference_name]))
                if -1 not in [alignments[n][i] for n in alignments]
            ]
        )
        reference_pdb = pd.parsePDB(
            str(
                self.output_folder
                / f"cleaned_pdb/{self.structures[self.reference_structure_index].name}.pdb"
            )
        )
        aln_ref = alignments[reference_name]
        ref_coords_core = (
            reference_pdb[helper.get_alpha_indices(reference_pdb)]
            .getCoords()
            .astype(np.float64)[np.array([aln_ref[c] for c in core_indices])]
        )
        ref_centroid = helper.nb_mean_axis_0(ref_coords_core)
        ref_coords_core -= ref_centroid
        transformation = pd.Transformation(np.eye(3), -ref_centroid)
        reference_pdb = pd.applyTransformation(transformation, reference_pdb)
        pd.writePDB(str(output_pdb_folder / f"{reference_name}.pdb"), reference_pdb)
        for i in range(1, len(self.structures)):
            name = self.structures[i].name
            pdb = pd.parsePDB(
                str(self.output_folder / f"cleaned_pdb/{self.structures[i].name}.pdb")
            )
            aln_name = alignments[name]
            common_coords_2 = (
                pdb[helper.get_alpha_indices(pdb)]
                .getCoords()
                .astype(np.float64)[np.array([aln_name[c] for c in core_indices])]
            )
            (
                rotation_matrix,
                translation_matrix,
            ) = superposition_functions.svd_superimpose(
                ref_coords_core, common_coords_2
            )
            transformation = pd.Transformation(rotation_matrix.T, translation_matrix)
            pdb = pd.applyTransformation(transformation, pdb)
            pd.writePDB(str(output_pdb_folder / f"{name}.pdb"), pdb)

    def write_superposed_pdbs_reference(self, output_pdb_folder, alignments):
        """
        Superposes PDBs according to reference structure and writes transformed PDBs to files
        (View with Pymol)

        Parameters
        ----------
        alignments
        output_pdb_folder
        """
        reference_name = self.structures[self.reference_structure_index].name
        reference_pdb = pd.parsePDB(
            str(
                self.output_folder
                / f"cleaned_pdb/{self.structures[self.reference_structure_index].name}.pdb"
            )
        )
        aln_ref = alignments[reference_name]
        reference_coords = (
            reference_pdb[helper.get_alpha_indices(reference_pdb)]
            .getCoords()
            .astype(np.float64)
        )
        pd.writePDB(str(output_pdb_folder / f"{reference_name}.pdb"), reference_pdb)
        for i in range(1, len(self.structures)):
            name = self.structures[i].name
            pdb = pd.parsePDB(
                str(self.output_folder / f"cleaned_pdb/{self.structures[i].name}.pdb")
            )
            aln_name = alignments[name]
            common_coords_1, common_coords_2 = get_common_coordinates(
                reference_coords,
                pdb[helper.get_alpha_indices(pdb)].getCoords().astype(np.float64),
                aln_ref,
                aln_name,
            )
            (
                rotation_matrix,
                translation_matrix,
            ) = superposition_functions.svd_superimpose(
                common_coords_1, common_coords_2
            )
            transformation = pd.Transformation(rotation_matrix.T, translation_matrix)
            pdb = pd.applyTransformation(transformation, pdb)
            pd.writePDB(str(output_pdb_folder / f"{name}.pdb"), pdb)

    def get_aligned_features(
        self, dssp_dir, num_threads, alignment: dict = None, only_dssp: bool = True
    ) -> typing.Dict[str, np.ndarray]:
        """
        Get dict of aligned features
        """
        if alignment is None:
            alignment = self.make_sequence_alignment()

        pdb_files = [
            self.output_folder / "cleaned_pdb" / f"{s.name}.pdb"
            for s in self.structures
        ]
        features = feature_extraction.get_features_multiple(
            pdb_files,
            str(dssp_dir),
            num_threads=num_threads,
            only_dssp=only_dssp,
            force_overwrite=True,
        )
        feature_names = list(features[0].keys())
        aligned_features = {}
        alignment_length = len(alignment[self.structures[0].name])
        for feature_name in feature_names:
            if feature_name == "secondary":
                continue
            aligned_features[feature_name] = np.zeros(
                (len(self.structures), alignment_length)
            )
            aligned_features[feature_name][:] = np.nan
            for p in range(len(self.structures)):
                farray = features[p][feature_name]
                if "gnm" in feature_name or "anm" in feature_name:
                    farray = farray / np.nansum(farray ** 2) ** 0.5
                indices = [
                    i
                    for i in range(alignment_length)
                    if alignment[self.structures[p].name][i] != "-"
                ]
                aligned_features[feature_name][p, indices] = farray
        return aligned_features

    def superpose(self, alignments: dict = None):
        """
        Superposes structures to first structure using Kabsch superposition
        """
        if alignments is None:
            alignments = self.alignment
        reference_key = self.structures[self.reference_structure_index].name
        core_indices = np.array(
            [
                i
                for i in range(len(alignments[reference_key]))
                if "-" not in [alignments[n][i] for n in alignments]
            ]
        )
        if len(core_indices) < len(alignments[reference_key]) // 2:
            self.superpose_reference(alignments)
        else:
            self.superpose_core(alignments, core_indices)

    def superpose_core(self, alignments: dict = None, core_indices: np.ndarray = None):
        """
        Superposes structures to first structure according to core positions in alignment using Kabsch superposition
        """
        if alignments is None:
            alignments = self.alignment
        reference_key = self.structures[self.reference_structure_index].name
        if core_indices is None:
            core_indices = np.array(
                [
                    i
                    for i in range(len(alignments[reference_key]))
                    if "-" not in [alignments[n][i] for n in alignments]
                ]
            )
        aln_ref = alignments[reference_key]
        ref_coords = self.structures[self.reference_structure_index].coordinates[
            np.array([aln_ref[c] for c in core_indices])
        ]
        ref_centroid = helper.nb_mean_axis_0(ref_coords)
        ref_coords -= ref_centroid
        for i in range(len(self.structures)):
            if i == self.reference_structure_index:
                self.structures[i].coordinates -= ref_centroid
            else:
                aln_c = alignments[self.structures[i].name]
                common_coords_2 = self.structures[i].coordinates[
                    np.array([aln_c[c] for c in core_indices])
                ]
                (
                    rotation_matrix,
                    translation_matrix,
                ) = superposition_functions.paired_svd_superpose(
                    ref_coords, common_coords_2
                )
                self.structures[i].coordinates = superposition_functions.apply_rotran(
                    self.structures[i].coordinates, rotation_matrix, translation_matrix
                )

    def superpose_reference(self, alignments: dict = None):
        """
        Superposes structures to first structure according to reference structure using Kabsch superposition
        """
        if alignments is None:
            alignments = self.alignment
        reference_key = self.structures[self.reference_structure_index].name
        aln_ref = alignments[reference_key]
        for i in range(len(self.structures)):
            aln_c = alignments[self.structures[i].name]
            common_coords_1, common_coords_2 = get_common_coordinates(
                self.structures[self.reference_structure_index].coordinates,
                self.structures[i].coordinates,
                aln_ref,
                aln_c,
            )
            assert common_coords_1.shape[0] > 0
            (
                rotation_matrix,
                translation_matrix,
            ) = superposition_functions.paired_svd_superpose(
                common_coords_1, common_coords_2
            )
            self.structures[i].coordinates = superposition_functions.apply_rotran(
                self.structures[i].coordinates, rotation_matrix, translation_matrix
            )

    def make_rmsd_coverage_tm_matrix(
        self, alignments: dict = None, superpose_first: bool = True
    ):
        """
        Find RMSDs and coverages of the alignment of each pair of sequences

        Parameters
        ----------
        alignments
            if None uses self.alignment
        superpose_first
            if True then superposes all structures to first structure first

        Returns
        -------
        RMSD matrix, coverage matrix
        """
        if alignments is None:
            alignments = self.alignment
        num = len(self.structures)
        pairwise_rmsd_matrix = np.zeros((num, num))
        pairwise_rmsd_matrix[:] = np.nan
        pairwise_coverage = np.zeros((num, num))
        pairwise_coverage[:] = np.nan
        pairwise_tm = np.zeros((num, num))
        pairwise_tm[:] = np.nan
        if superpose_first:
            self.superpose(alignments)
        for i in range(num - 1):
            for j in range(i + 1, num):
                name_1, name_2 = self.structures[i].name, self.structures[j].name
                aln_1 = alignments[name_1]
                aln_2 = alignments[name_2]
                common_coords_1, common_coords_2 = get_common_coordinates(
                    self.structures[i].coordinates,
                    self.structures[j].coordinates,
                    aln_1,
                    aln_2,
                )
                assert common_coords_1.shape[0] > 0
                if not superpose_first:
                    rot, tran = superposition_functions.paired_svd_superpose(
                        common_coords_1, common_coords_2
                    )
                    common_coords_2 = superposition_functions.apply_rotran(
                        common_coords_2, rot, tran
                    )
                pairwise_rmsd_matrix[i, j] = pairwise_rmsd_matrix[
                    j, i
                ] = score_functions.get_rmsd(common_coords_1, common_coords_2)
                pairwise_coverage[i, j] = pairwise_coverage[
                    j, i
                ] = common_coords_1.shape[0] / len(aln_1)
                pairwise_tm[i, j] = pairwise_tm[j, i] = tm_score(
                    common_coords_1,
                    common_coords_2,
                    self.structures[i].length,
                    self.structures[j].length,
                )
        return pairwise_rmsd_matrix, pairwise_coverage, pairwise_tm

    def get_aligned_structures(self, alignment=None):
        if alignment is None:
            alignment = self.alignment
        if type(alignment[self.structures[0].name]) == str:
            alignment = alignment_to_numpy(alignment)
        self.superpose(alignment)
        aligned_structures = np.zeros(
            (len(self.structures), len(alignment[self.structures[0].name]), 3)
        )
        nan_3 = np.zeros(3)
        nan_3[:] = np.nan
        for i in range(len(self.structures)):
            aligned_structures[i] = [
                self.structures[i].coordinates[x] if x != -1 else nan_3
                for x in alignment[self.structures[i].name]
            ]
        return aligned_structures

    def get_profile_alignment(
        self, msa_class_new, gap_open_penalty: float, gap_extend_penalty: float,
    ):
        # Assumes self has already been aligned
        msa_class_profile = deepcopy(self)
        profile_sequence = np.arange(
            len(msa_class_profile.alignment[msa_class_profile.structures[0].name])
        )
        aligned_structures = msa_class_profile.get_aligned_structures()
        profile_coords = np.nanmean(
            np.array(
                [
                    az.hdi(aligned_structures[:, i], skipna=True)
                    for i in range(aligned_structures.shape[1])
                ]
            ),
            axis=-1,
        )
        profile_structure = Structure(
            "caretta_profile", aligned_structures.shape[1], profile_coords
        )
        alignment = msa_class_profile.make_sequence_alignment()
        for new_structure in msa_class_new.structures:
            msa_class = StructureMultiple(
                [profile_structure, new_structure],
                {
                    "caretta_profile": profile_sequence,
                    new_structure.name: msa_class_new.sequences[new_structure.name],
                },
                msa_class_profile.superposition_parameters,
                msa_class_profile.superposition_function,
            )
            alignment_indices = msa_class.align(
                gap_open_penalty, gap_extend_penalty, False
            )
            alignment = {}
            alignment[new_structure.name] = "".join(
                msa_class_new.sequences[new_structure.name][i] if i != -1 else "-"
                for i in alignment_indices[new_structure.name]
            )
            for n in msa_class_profile.sequences:
                if n != new_structure.name:
                    alignment[n] = "".join(
                        msa_class_profile.sequences[n][
                            msa_class_profile.alignment[n][i]
                        ]
                        if i != -1 and msa_class_profile.alignment[n][i] != -1
                        else "-"
                        for i in alignment_indices["caretta_profile"]
                    )
            msa_class_profile.structures.append(new_structure)
            msa_class_profile.sequences[new_structure.name] = msa_class_new.sequences[
                new_structure.name
            ]
            msa_class_profile.alignment = alignment_to_numpy(alignment)
            aligned_structures = msa_class_profile.get_aligned_structures()
            profile_coords = np.nanmean(
                np.array(
                    [
                        az.hdi(aligned_structures[:, i], skipna=True)
                        for i in range(aligned_structures.shape[1])
                    ]
                ),
                axis=-1,
            )
            profile_structure = Structure(
                "caretta_profile", aligned_structures.shape[1], profile_coords
            )
        return alignment


def trigger_numba_compilation():
    """
    Run this at the beginning of a Caretta run to compile Numba functions
    """
    parameters = {
        "size": 1,
        "gap_open_penalty": 0.0,
        "gap_extend_penalty": 0.0,
        "gamma": 0.03,
    }
    coords_1 = np.zeros((2, 3))
    coords_2 = np.zeros((2, 3))
    tm_score(coords_1, coords_2, 2, 2)
    weights_1 = np.zeros((coords_1.shape[0], 1))
    weights_2 = np.zeros((coords_2.shape[0], 1))
    get_pairwise_alignment(
        coords_1, coords_2, parameters["gamma"], 0, 0, weights_1, weights_2
    )
    superposition_functions.signal_svd_superpose_function(
        coords_1, coords_2, parameters
    )
    distance_matrix = np.random.random((5, 5))
    nj.neighbor_joining(distance_matrix)
    aln_1 = np.array([0, -1, 1])
    aln_2 = np.array([0, 1, -1])
    get_common_coordinates(coords_1, coords_2, aln_1, aln_2)
    get_mean_coords(aln_1, coords_1, aln_2, coords_2)



import dash_core_components as dcc
import dash_html_components as html
import numpy as np
from caretta.app import app_helper
from pathlib import Path

box_style = {
    "box-shadow": "1px 3px 20px -4px rgba(0,0,0,0.75)",
    "border-radius": "5px",
    "background-color": "#f9f7f7",
}


def get_layout(
    introduction_text,
    input_text,
    placeholder_text,
    selection_text,
    suite,
    pfam_class=None,
):
    return html.Div(
        children=[
            get_introduction_panel(introduction_text),
            html.Br(),
            get_input_panel_layout(
                input_text, placeholder_text, selection_text, pfam_class
            ),
            html.Br(),
            get_hidden_variables_layout(suite),
            get_sequence_alignment_layout(),
            html.Br(),
            get_structure_alignment_layout(),
            html.Br(),
            get_feature_alignment_panel(),
            html.Br(),
        ]
    )


def get_introduction_panel(introduction_text: str):
    return html.Div(
        html.Div(
            [
                html.H1("Caretta", style={"text-align": "center"}),
                html.H3(
                    "a multiple protein structure alignment and feature extraction suite",
                    style={"text-align": "center"},
                ),
                html.P(dcc.Markdown(introduction_text), style={"text-align": "left"}),
            ],
            className="row",
        ),
        className="container",
    )


def get_hidden_variables_layout(suite):
    return html.Div(
        [
            html.P(children="", id="proteins-list", style={"display": "none"}),
            html.P(children="", id="feature-alignment-data", style={"display": "none"}),
            html.P(
                children=app_helper.compress_object(0, suite),
                id="structure-alignment-selected-residue",
                style={"display": "none"},
            ),
            html.P(
                children=app_helper.compress_object(0, suite),
                id="feature-alignment-selected-residue",
                style={"display": "none"},
            ),
            html.P(
                children="", id="sequence-alignment-data", style={"display": "none"}
            ),
            html.P(children="", id="caretta-class", style={"display": "none"}),
            html.P(
                children=app_helper.compress_object(
                    np.random.randint(0, 1000000000), suite
                ),
                id="unique-id",
                style={"display": "none"},
            ),
        ]
    )


def get_input_panel_layout(
    input_text: str,
    placeholder_text: str,
    selection_text: str,
    pfam_class: app_helper.PfamToPDB = None,
):
    if pfam_class is not None:
        user_input = dcc.Dropdown(
            placeholder="Choose Pfam ID",
            options=[{"label": x, "value": x} for x in pfam_class.pfam_to_pdb_ids],
            id="user-input",
        )

    else:
        user_input = (
            dcc.Textarea(
                placeholder=placeholder_text, value="", id="user-input", required=True,
            ),
        )
    return html.Div(
        [
            html.Div(
                [
                    html.Br(),
                    html.H3(
                        "Choose Structures",
                        className="row",
                        style={"text-align": "center"},
                    ),
                    html.P(input_text, className="row",),
                    html.Div(
                        [
                            html.Div(user_input, className="four columns",),
                            html.P(
                                dcc.Markdown(selection_text), className="four columns"
                            ),
                            html.Button(
                                "Load Structures",
                                className="four columns",
                                id="load-structures-button",
                            ),
                        ],
                        className="row",
                    ),
                    html.Div(
                        [
                            html.Div(
                                dcc.Dropdown(
                                    placeholder="Gap open penalty (1.0)",
                                    options=[
                                        {"label": np.round(x, decimals=2), "value": x,}
                                        for x in np.arange(0, 5, 0.1)
                                    ],
                                    id="gap-open-dropdown",
                                ),
                                className="four columns",
                            ),
                            html.Div(
                                dcc.Dropdown(
                                    multi=True,
                                    id="proteins-selection-dropdown",
                                    placeholder="Select PDB IDs to align",
                                ),
                                className="four columns",
                            ),
                            html.Div(
                                dcc.Dropdown(
                                    placeholder="Gap extend penalty (0.01)",
                                    options=[
                                        {"label": np.round(x, decimals=3), "value": x,}
                                        for x in np.arange(0, 1, 0.002)
                                    ],
                                    id="gap-extend-dropdown",
                                ),
                                className="four columns",
                            ),
                        ],
                        className="row",
                    ),
                    html.Br(),
                    html.Div(
                        html.Button(
                            "Align Structures",
                            className="twelve columns",
                            id="align-button",
                        ),
                        className="row",
                    ),
                    dcc.Loading(
                        id="loading-indicator",
                        children=[
                            html.Div(
                                id="loading-indicator-output",
                                style={"text-align": "center"},
                            )
                        ],
                        type="default",
                    ),
                    html.P(
                        id="time-estimate",
                        style={"text-align": "center"},
                        children="",
                        className="row",
                    ),
                ],
                className="container",
            ),
            html.Br(),
        ],
        className="container",
        style=box_style,
    )


def get_sequence_alignment_layout():
    return html.Div(
        children=[
            html.Br(),
            html.H3(
                "Sequence alignment", className="row", style={"text-align": "center"},
            ),
            html.Div(html.P("", className="row"), className="container"),
            html.Div(
                [
                    html.Button(
                        "Download sequence alignment",
                        className="twelve columns",
                        id="fasta-download-button",
                    ),
                    html.Div(children="", className="row", id="fasta-download-link"),
                ],
                className="container",
            ),
            html.Div(
                html.P(id="sequence-alignment", className="twelve columns"),
                className="row",
            ),
        ],
        className="container",
        style=box_style,
    )


def get_structure_alignment_layout():
    return html.Div(
        [
            html.Br(),
            html.H3(
                "Structural alignment", className="row", style={"text-align": "center"},
            ),
            html.Div(
                html.P(
                    "Click on a residue to see its position on the feature alignment in the next section.",
                    className="row",
                ),
                className="container",
            ),
            html.Div(
                [
                    html.Button(
                        "Download superposed PDBs",
                        id="pdb-download-button",
                        className="twelve columns",
                    ),
                    html.Div(children="", className="row", id="pdb-download-link"),
                ],
                className="container",
            ),
            html.Div(
                children=dcc.Graph(figure=app_helper.empty_dict(), id="scatter-plot",),
                className="row",
                id="structure-alignment",
            ),
            html.Br(),
        ],
        className="container",
        style=box_style,
    )


def get_feature_alignment_panel():
    return html.Div(
        [
            html.Br(),
            html.Div(
                [
                    html.Div(
                        [
                            html.H3(
                                "Feature alignment",
                                className="row",
                                style={"text-align": "center"},
                            ),
                            html.P(
                                "Click on a position in the feature alignment to see the corresponding residues in the previous section.",
                                className="row",
                            ),
                            dcc.Dropdown(
                                placeholder="Choose a feature",
                                id="feature-selection-dropdown",
                                className="six columns",
                            ),
                            html.Button(
                                "Display feature alignment",
                                id="display-feature-button",
                                className="six columns",
                            ),
                        ],
                        className="row",
                    ),
                    html.Br(),
                    html.Div(
                        [
                            html.Div(
                                get_export_feature_buttons(), id="feature-exporter",
                            ),
                            html.Div(html.P(""), id="feature-download-link"),
                            html.Br(),
                        ]
                    ),
                ],
                className="container",
            ),
            html.Div(
                html.Div(
                    dcc.Graph(figure=app_helper.empty_dict(), id="feature-line-graph",),
                    id="feature-line",
                    className="twelve columns",
                ),
                className="row",
            ),
            html.Div(
                html.Div(
                    dcc.Graph(
                        figure=app_helper.empty_dict(), id="feature-heatmap-graph",
                    ),
                    id="feature-heatmap",
                    className="twelve columns",
                ),
                className="row",
            ),
        ],
        className="container",
        style=box_style,
    )


def get_export_feature_buttons():
    return [
        html.Div(
            html.Button(
                "Download feature as tab-separated file",
                id="export-feature-button",
                className="twelve columns",
            ),
            className="row",
        ),
        html.Br(),
        html.Div(
            html.Button(
                "Download all features",
                id="export-all-features-button",
                className="twelve columns",
            ),
            className="row",
        ),
        html.Br(),
    ]


def get_download_string(filename):
    return html.Div(
        html.A(
            f"Download {Path(filename).stem}{Path(filename).suffix} here",
            href=f"/caretta/{filename}",
            className="twelve columns",
        ),
        className="container",
    )



import numba as nb
import numpy as np
from caretta import helper


@nb.njit
def get_caretta_score(coord_1: np.ndarray, coord_2: np.ndarray, gamma=0.03):
    """
    Gaussian (RBF) score of similarity between two coordinates
    """
    return np.exp(-gamma * np.sum((coord_1 - coord_2) ** 2, axis=-1))


@nb.njit
def get_signal_score(signal_1, signal_2, gamma=0.1):
    """
    Used in superposition_functions.signal_superpose_function
    """
    return np.median(np.exp(-gamma * (signal_1 - signal_2) ** 2))


@nb.njit
def get_rmsd(coords_1: np.ndarray, coords_2: np.ndarray) -> float:
    """
    RMSD of paired coordinates = normalized square-root of sum of squares of euclidean distances
    """
    return np.sqrt(np.sum((coords_1 - coords_2) ** 2) / coords_1.shape[0])


@nb.njit
def make_score_matrix(
    coords_1: np.ndarray, coords_2: np.ndarray, score_function, gamma, normalized=False
) -> np.ndarray:
    """
    Makes matrix of scores of each coordinate in coords_1 to each coordinate in coords_2
    Parameters
    ----------
    coords_1
        shape = (n, 3)
    coords_2
        shape = (m, 3)
    score_function
    gamma
    normalized
    Returns
    -------
    matrix; shape = (n, m)
    """
    score_matrix = np.zeros((coords_1.shape[0], coords_2.shape[0]))

    if normalized:
        both = np.concatenate((coords_1, coords_2))
        mean, std = helper.nb_mean_axis_0(both), helper.nb_std_axis_0(both)
        coords_1 = (coords_1 - mean) / std
        coords_2 = (coords_2 - mean) / std
    for i in range(coords_1.shape[0]):
        for j in range(coords_2.shape[0]):
            score_matrix[i, j] = score_function(coords_1[i], coords_2[j], gamma)
    return score_matrix


@nb.njit
def get_total_score(
    coords_1: np.ndarray, coords_2: np.ndarray, score_function, gamma, normalized=False
) -> float:
    """
    Get total score for a set of paired coordinates

    Parameters
    ----------
    coords_1
    coords_2
    score_function
    gamma
    normalized
        if True, divides by length of first coordinate set

    Returns
    -------
    Total score
    """
    score = 0
    for i in range(coords_1.shape[0]):
        score += score_function(coords_1[i], coords_2[i], gamma)
    if normalized:
        return score / coords_1.shape[0]
    else:
        return score



from pathlib import Path
from zipfile import ZipFile

import dash
import dash_bio
import dash_core_components as dcc
import numpy as np
import pyparsing

from caretta import multiple_alignment
from caretta.app import app_helper, app_layout
from flask import send_from_directory, abort


def register_callbacks(app, get_pdb_entries, suite):
    """
    Register callbacks to dash app

    Parameters
    ----------
    app
        Dash app
    get_pdb_entries
        function that takes a single user inputted string and returns a list of PdbEntry objects
    suite
        Fernet object
    """
    # Function called when user inputs folder / Pfam ID and clicks "Load Structures"
    @app.callback(
        # Output
        # PDB files / PDB IDs found in user-inputted folder / Pfam ID
        dash.dependencies.Output("proteins-selection-dropdown", "options"),
        # Input
        [
            # Whether the load-structures-button has been clicked
            dash.dependencies.Input("load-structures-button", "n_clicks")
        ],
        # State
        [
            # User input (PDB folder / Pfam ID)
            dash.dependencies.State("user-input", "value"),
        ],
    )
    def load_structures(load_structures_button, user_input):
        if load_structures_button and user_input:
            pdb_entries = get_pdb_entries(user_input)
            labels = []
            for x in pdb_entries:
                label = x.PDB_ID
                if x.CHAIN_ID != "none":
                    label += f".{x.CHAIN_ID}"
                if x.PdbResNumStart != -1 and x.PdbResNumEnd != -1:
                    label += f" {x.PdbResNumStart}-{x.PdbResNumEnd}"
                labels.append(label)
            return [
                {"label": label, "value": app_helper.compress_object(x, suite)}
                for label, x in zip(labels, pdb_entries)
            ]
        else:
            return [{"label": "no selection", "value": "None"}]

    # Function called when user selects IDs to align and clicks Align
    @app.callback(
        # Outputs
        [
            # hidden Divs
            # MSA class
            dash.dependencies.Output("caretta-class", "children"),
            # aligned sequences dict
            dash.dependencies.Output("sequence-alignment-data", "children"),
            # aligned feature dict
            dash.dependencies.Output("feature-alignment-data", "children"),
            # sequence alignment panel
            dash.dependencies.Output("sequence-alignment", "children"),
            # structure alignment panel
            dash.dependencies.Output("structure-alignment", "children"),
            # feature panel selection box
            dash.dependencies.Output("feature-selection-dropdown", "options"),
            # stop loading indicator
            dash.dependencies.Output("loading-indicator-output", "children"),
        ],
        # Inputs
        [
            # Whether the Align button was clicked
            dash.dependencies.Input("align-button", "n_clicks")
        ],
        # States
        [
            # User input (PDB folder / Pfam ID)
            dash.dependencies.State("user-input", "value"),
            # Selected PDB IDs
            dash.dependencies.State("proteins-selection-dropdown", "value"),
            # Penalties
            dash.dependencies.State("gap-open-dropdown", "value"),
            dash.dependencies.State("gap-extend-dropdown", "value"),
            # Unique ID for setting output folder
            dash.dependencies.State("unique-id", "children"),
        ],
    )
    def align_structures(
        align_button,
        user_input,
        proteins_selection_dropdown,
        gap_open_dropdown,
        gap_extend_dropdown,
        unique_id,
    ):
        if align_button and user_input and proteins_selection_dropdown:
            pdb_entries = [
                app_helper.decompress_object(x, suite)
                for x in proteins_selection_dropdown
            ]
            if not gap_open_dropdown:
                gap_open_dropdown = 1
            if not gap_extend_dropdown:
                gap_extend_dropdown = 0.01
            pdb_files = []
            for p in pdb_entries:
                try:
                    pdb_files.append(p.get_pdb()[1])
                except (OSError, AttributeError, pyparsing.ParseException):
                    continue
            msa_class = multiple_alignment.StructureMultiple.from_pdb_files(
                pdb_files,
                multiple_alignment.DEFAULT_SUPERPOSITION_PARAMETERS,
                output_folder=f"static/results_{app_helper.decompress_object(unique_id, suite)}",
            )
            if len(msa_class.structures) > 2:
                msa_class.make_pairwise_shape_matrix()
                sequence_alignment = msa_class.align(
                    gap_open_penalty=gap_open_dropdown,
                    gap_extend_penalty=gap_extend_dropdown,
                )
            else:
                sequence_alignment = msa_class.align(
                    gap_open_penalty=gap_open_dropdown,
                    gap_extend_penalty=gap_extend_dropdown,
                )
            msa_class.superpose()
            fasta = app_helper.to_fasta_str(sequence_alignment)
            dssp_dir = msa_class.output_folder / ".caretta_tmp"
            if not dssp_dir.exists():
                dssp_dir.mkdir()
            features = msa_class.get_aligned_features(
                dssp_dir, num_threads=4, only_dssp=False
            )
            caretta_class = app_helper.compress_object(msa_class, suite)
            sequence_alignment_data = app_helper.compress_object(
                sequence_alignment, suite
            )
            feature_alignment_data = app_helper.compress_object(features, suite)

            sequence_alignment_component = dash_bio.AlignmentChart(
                id="sequence-alignment-graph",
                data=fasta,
                showconsensus=False,
                showconservation=False,
                overview=None,
                height=300,
                colorscale="hydrophobicity",
            )
            structure_alignment_component = dcc.Graph(
                figure=app_helper.scatter3D(
                    {s.name: s.coordinates for s in msa_class.structures}
                ),
                id="scatter-plot",
            )
            feature_selection_dropdown = [{"label": x, "value": x} for x in features]
            loading_indicator_output = ""
            return (
                caretta_class,
                sequence_alignment_data,
                feature_alignment_data,
                sequence_alignment_component,
                structure_alignment_component,
                feature_selection_dropdown,
                loading_indicator_output,
            )
        else:
            return (
                app_helper.empty_object(suite),
                app_helper.empty_object(suite),
                app_helper.empty_object(suite),
                "",
                "",
                [{"label": "no alignment present", "value": "no alignment"}],
                "",
            )

    # Function that displays mean +/- stdev and heatmap of user-selected feature
    @app.callback(
        # Outputs
        [
            # Feature line panel
            dash.dependencies.Output("feature-line", "children"),
            # Feature heatmap panel
            dash.dependencies.Output("feature-heatmap", "children"),
        ],
        # Inputs
        [
            # Whether the display feature button has been clicked
            dash.dependencies.Input("display-feature-button", "n_clicks")
        ],
        # States
        [
            # Dropdown of feature names
            dash.dependencies.State("feature-selection-dropdown", "value"),
            # Aligned features dict
            dash.dependencies.State("feature-alignment-data", "children"),
        ],
    )
    def display_feature(
        display_feature_button_clicked,
        feature_selection_dropdown_value,
        feature_alignment_data,
    ):
        if (
            display_feature_button_clicked
            and feature_selection_dropdown_value
            and feature_alignment_data
        ):
            feature_alignment_dict = app_helper.decompress_object(
                feature_alignment_data, suite
            )
            chosen_feature_data = feature_alignment_dict[
                feature_selection_dropdown_value
            ]
            feature_line_component = dcc.Graph(
                figure=app_helper.line(chosen_feature_data), id="feature-line-graph"
            )
            feature_heatmap_component = dcc.Graph(
                figure=app_helper.heatmap(chosen_feature_data),
                id="feature-heatmap-graph",
            )
            return feature_line_component, feature_heatmap_component
        else:
            return (
                dcc.Graph(
                    figure=app_helper.empty_dict(),
                    id="feature-line-graph",
                    style={"display": "none"},
                ),
                dcc.Graph(
                    figure=app_helper.empty_dict(),
                    id="feature-heatmap-graph",
                    style={"display": "none"},
                ),
            )

    # Function that updated selected residues in structure alignment panel and feature alignment panel
    @app.callback(
        # Outputs
        [
            # hidden Divs
            # Residue position selected in structure alignment panel
            dash.dependencies.Output(
                "structure-alignment-selected-residue", "children"
            ),
            # Residue position selected in feature alignment panel
            dash.dependencies.Output("feature-alignment-selected-residue", "children"),
            # Feature line component
            dash.dependencies.Output("feature-line-graph", "figure"),
            # 3D scatter component
            dash.dependencies.Output("scatter-plot", "figure"),
        ],
        # Inputs
        [
            # Clicked indices in 3D scatter component
            dash.dependencies.Input("scatter-plot", "clickData"),
            # Clicked indices in feature line component
            dash.dependencies.Input("feature-line-graph", "clickData"),
        ],
        # States
        [
            # Feature line component
            dash.dependencies.State("feature-line-graph", "figure"),
            # 3D scatter component
            dash.dependencies.State("scatter-plot", "figure"),
            # Residue position selected in structure alignment panel
            dash.dependencies.State("structure-alignment-selected-residue", "children"),
            # Residue position selected in feature alignment panel
            dash.dependencies.State("feature-alignment-selected-residue", "children"),
            # Aligned sequences dict
            dash.dependencies.State("sequence-alignment-data", "children"),
        ],
    )
    def update_interactive_panels(
        scatter_plot_clickdata,
        feature_line_clickdata,
        feature_line_graph,
        scatter_plot,
        structure_alignment_selected_residue,
        feature_alignment_selected_residue,
        sequence_alignment_data,
    ):
        if feature_line_graph and scatter_plot:
            changed = None
            clickdata = None
            if (
                feature_line_clickdata
                and app_helper.compress_object(
                    (
                        feature_line_clickdata["points"][0]["pointNumber"],
                        feature_line_clickdata["points"][0]["curveNumber"],
                    ),
                    suite,
                )
                != feature_alignment_selected_residue
            ):
                clickdata = feature_line_clickdata
                changed = "feature-panel"
            elif (
                scatter_plot_clickdata
                and app_helper.compress_object(
                    (
                        scatter_plot_clickdata["points"][0]["pointNumber"],
                        scatter_plot_clickdata["points"][0]["curveNumber"],
                    ),
                    suite,
                )
                != structure_alignment_selected_residue
            ):
                clickdata = scatter_plot_clickdata
                changed = "structure-panel"
            if changed is not None and clickdata is not None:
                # Save new clicked index
                aln_index = clickdata["points"][0]["pointNumber"]
                protein_index = clickdata["points"][0]["curveNumber"]
                if changed == "feature-panel":
                    feature_alignment_selected_residue = app_helper.compress_object(
                        (aln_index, protein_index), suite
                    )
                elif changed == "structure-panel":
                    structure_alignment_selected_residue = app_helper.compress_object(
                        (aln_index, protein_index), suite
                    )

                sequence_alignment = app_helper.decompress_object(
                    sequence_alignment_data, suite
                )
                number_of_structures = len(sequence_alignment)

                try:
                    maxim, minim = (
                        np.max(feature_line_graph["data"][0]["y"]),
                        np.min(feature_line_graph["data"][0]["y"]),
                    )
                except KeyError:
                    return (
                        structure_alignment_selected_residue,
                        feature_alignment_selected_residue,
                        feature_line_graph,
                        scatter_plot,
                    )
                if len(feature_line_graph["data"]) > 2:
                    feature_line_graph["data"] = feature_line_graph["data"][:-1]
                if len(scatter_plot["data"]) > number_of_structures:
                    scatter_plot["data"] = scatter_plot["data"][:-1]

                if changed == "feature-panel":
                    aln_positions = app_helper.aln_index_to_protein(
                        aln_index, sequence_alignment
                    )
                    feature_line_graph["data"] += [
                        dict(
                            y=[minim, maxim],
                            x=[aln_index, aln_index],
                            type="scatter",
                            mode="lines",
                            name="selected residue",
                        )
                    ]

                    to_add = []
                    for i in range(len(scatter_plot["data"])):
                        p = aln_positions[scatter_plot["data"][i]["name"]]
                        if p is not None:
                            x, y, z = (
                                scatter_plot["data"][i]["x"][p],
                                scatter_plot["data"][i]["y"][p],
                                scatter_plot["data"][i]["z"][p],
                            )
                            to_add.append((x, y, z))
                        else:
                            continue
                    scatter_plot["data"] += [
                        dict(
                            x=[x[0] for x in to_add],
                            y=[y[1] for y in to_add],
                            z=[z[2] for z in to_add],
                            type="scatter3d",
                            mode="markers",
                            name="selected residues",
                        )
                    ]
                elif changed == "structure-panel":
                    aligned_sequence = list(sequence_alignment.values())[protein_index]
                    aln_index = app_helper.protein_to_aln_index(
                        aln_index, aligned_sequence
                    )
                    x, y, z = (
                        clickdata["points"][0]["x"],
                        clickdata["points"][0]["y"],
                        clickdata["points"][0]["z"],
                    )
                    feature_line_graph["data"] += [
                        dict(
                            y=[minim, maxim],
                            x=[aln_index, aln_index],
                            type="scatter",
                            mode="lines",
                            name="selected_residue",
                        )
                    ]
                    scatter_plot["data"] += [
                        dict(
                            y=[y],
                            x=[x],
                            z=[z],
                            type="scatter3d",
                            mode="markers",
                            name="selected residue",
                        )
                    ]
        return (
            structure_alignment_selected_residue,
            feature_alignment_selected_residue,
            feature_line_graph,
            scatter_plot,
        )

    # Function to export FASTA alignment file
    @app.callback(
        # Outputs
        # Link ot download file
        dash.dependencies.Output("fasta-download-link", "children"),
        # Inputs
        [
            # Whether the FASTA download button has been clicked
            dash.dependencies.Input("fasta-download-button", "n_clicks")
        ],
        # States
        [
            # Aligned sequences dict
            dash.dependencies.State("sequence-alignment-data", "children"),
            # MSA class
            dash.dependencies.State("caretta-class", "children"),
        ],
    )
    def download_alignment(
        fasta_download_button_clicked, sequence_alignment_data, caretta_class
    ):
        if fasta_download_button_clicked and sequence_alignment_data and caretta_class:
            sequence_alignment = app_helper.decompress_object(
                sequence_alignment_data, suite
            )
            if not sequence_alignment:
                return ""
            msa_class = app_helper.decompress_object(caretta_class, suite)
            msa_class.write_files(
                write_pdb=False,
                write_fasta=True,
                write_class=False,
                write_features=False,
                write_matrix=False,
            )
            return app_layout.get_download_string(
                str(Path(msa_class.output_folder) / "result.fasta")
            )
        else:
            return ""

    # Function to export superposed PDB files
    @app.callback(
        # Output
        # Link to download files
        dash.dependencies.Output("pdb-download-link", "children"),
        # Inputs
        [
            # Whether the PDB download button has been clicked
            dash.dependencies.Input("pdb-download-button", "n_clicks")
        ],
        # States
        [
            # Aligned sequences dict
            dash.dependencies.State("sequence-alignment-data", "children"),
            # MSA class
            dash.dependencies.State("caretta-class", "children"),
        ],
    )
    def download_pdb(
        pdb_download_button_clicked, sequence_alignment_data, caretta_class
    ):
        if pdb_download_button_clicked and sequence_alignment_data and caretta_class:
            sequence_alignment = app_helper.decompress_object(
                sequence_alignment_data, suite
            )
            if not sequence_alignment:
                return ""
            msa_class = app_helper.decompress_object(caretta_class, suite)

            msa_class.write_files(
                write_pdb=True,
                write_fasta=False,
                write_class=False,
                write_features=False,
                write_matrix=False,
            )
            output_filename = f"{msa_class.output_folder}/superposed_pdbs.zip"
            pdb_zip_file = ZipFile(output_filename, mode="w")
            for pdb_file in (Path(msa_class.output_folder) / "superposed_pdbs").glob(
                "*.pdb"
            ):
                pdb_zip_file.write(
                    str(pdb_file), arcname=f"{pdb_file.stem}{pdb_file.suffix}"
                )
            return app_layout.get_download_string(output_filename)
        else:
            return ""

    # Function to export feature files
    @app.callback(
        # Outputs
        [
            # Link to download feature files
            dash.dependencies.Output("feature-download-link", "children"),
            # # Div with export buttons
            dash.dependencies.Output("feature-exporter", "children"),
        ],
        # Inputs
        [
            # Whether the export feature button has been clicked
            dash.dependencies.Input("export-feature-button", "n_clicks"),
            # Whether the export all features button has been clicked
            dash.dependencies.Input("export-all-features-button", "n_clicks"),
        ],
        # States
        [
            # Selected value in the dropdown of feature names
            dash.dependencies.State("feature-selection-dropdown", "value"),
            # Aligned feature dict
            dash.dependencies.State("feature-alignment-data", "children"),
            # MSA class
            dash.dependencies.State("caretta-class", "children"),
        ],
    )
    def download_features(
        export_feature_button_clicked,
        export_all_features_button_clicked,
        feature_selection_dropdown_value,
        feature_alignment_data,
        caretta_class,
    ):
        output_string = ""
        if feature_alignment_data and caretta_class:
            feature_alignment_dict = app_helper.decompress_object(
                feature_alignment_data, suite
            )
            msa_class = app_helper.decompress_object(caretta_class, suite)
            protein_names = [s.name for s in msa_class.structures]
            if (
                export_feature_button_clicked and feature_selection_dropdown_value
            ) and not export_all_features_button_clicked:
                output_filename = f"{msa_class.output_folder}/{'-'.join(feature_selection_dropdown_value.split())}.csv"
                app_helper.write_feature_as_tsv(
                    feature_alignment_dict[feature_selection_dropdown_value],
                    protein_names,
                    output_filename,
                )
                output_string = app_layout.get_download_string(output_filename)
            elif (
                export_all_features_button_clicked and not export_feature_button_clicked
            ):
                output_filename = f"{msa_class.output_folder}/features.zip"
                features_zip_file = ZipFile(output_filename, mode="w")
                for feature in feature_alignment_dict:
                    feature_file = (
                        f"{msa_class.output_folder}/{'-'.join(feature.split())}.csv"
                    )
                    app_helper.write_feature_as_tsv(
                        feature_alignment_dict[feature], protein_names, feature_file
                    )
                    features_zip_file.write(
                        str(feature_file), arcname=f"{'-'.join(feature.split())}.csv"
                    )
                output_string = app_layout.get_download_string(output_filename)
        return output_string, app_layout.get_export_feature_buttons()

    @app.server.route("/caretta/static/<path:path>")
    def download(path):
        """Serve a file from the static directory."""
        try:
            return send_from_directory(str(Path.cwd() / "static"), path)
        except FileNotFoundError:
            abort(404)



import numba as nb
import numpy as np


# Neighbor joining takes as input a distance matrix specifying the distance between each pair of taxa.
# The algorithm starts with a completely unresolved tree, whose topology corresponds to that of a star network,
# and iterates over the following steps until the tree is completely resolved and all branch lengths are known:
#     1. Based on the current distance matrix calculate the matrix Q (defined below).
#     2. Find the pair of distinct taxa i and j (i.e. with i ≠ j) for which Q(i,j) has its lowest value.
#        These taxa are joined to a newly created node, which is connected to the central node
#     3. Calculate the distance from each of the taxa in the pair to this new node (branch lengths).
#     4. Calculate the distance from each of the taxa outside of this pair to the new node.
#     5. Start the algorithm again, replacing the pair of joined neighbors with the new node and using the distances calculated in the previous step.


@nb.njit
# @numba_cc.export('neighbor_joining', '(f64[:])')
def neighbor_joining(distance_matrix: np.ndarray) -> (np.ndarray, np.ndarray):
    """
    Runs the neighbor joining algorithm on a distance matrix
    Returns guide tree as adjacency list + branch lengths

    Parameters
    ----------
    distance_matrix

    Returns
    -------
    tree (node_1, node_2) pairs
    branch_lengths corresponding to each tree pair
    (node indices above length of the distance_matrix correspond to intermediate nodes)
    """
    length = n = distance_matrix.shape[0]
    tree = np.zeros((length ** 2, 2), dtype=np.uint64)
    branch_lengths = np.zeros((length ** 2, 1), dtype=np.float64)
    index = 0
    true_indices = np.array(list(range(length)))
    num_intermediate_nodes = 0
    while n > 3:
        # indices of nodes to be joined (according to the current distance_matrix, not the initial one!)
        min_ij = _find_join_nodes(distance_matrix)
        # branch lengths of each node being joined to the new node created after they are joined
        delta_ij_u = _find_branch_length(distance_matrix, min_ij[0], min_ij[1])

        # make an intermediate node
        intermediate_node = num_intermediate_nodes + length
        num_intermediate_nodes += 1

        # add to tree
        tree[index] = np.array((true_indices[min_ij[0]], intermediate_node))
        branch_lengths[index] = delta_ij_u[0]
        index += 1
        tree[index] = np.array((true_indices[min_ij[1]], intermediate_node))
        branch_lengths[index] = delta_ij_u[1]
        index += 1

        # Distances of remaining indices to newly created node (step 4)
        indices = np.array([i for i in range(n) if i != min_ij[0] and i != min_ij[1]])
        new_distance_matrix = np.zeros((n - 1, n - 1))
        new_distance_matrix[1:, 1:] = distance_matrix[indices, :][:, indices]
        for i in range(len(indices)):
            new_distance_matrix[0, i + 1] = new_distance_matrix[i + 1, 0] = 0.5 * (
                distance_matrix[min_ij[0], indices[i]]
                + distance_matrix[min_ij[1], indices[i]]
                - distance_matrix[min_ij[0], min_ij[1]]
            )

        # Repeat (step 5)
        distance_matrix = new_distance_matrix
        n = distance_matrix.shape[0]
        true_indices = np.array(
            [intermediate_node] + [true_indices[i] for i in indices]
        )

    # Last 3 nodes
    delta_ij_u = _find_branch_length(distance_matrix, 1, 2)
    intermediate_node = num_intermediate_nodes + length
    num_intermediate_nodes += 1

    tree[index] = np.array((true_indices[1], intermediate_node))
    branch_lengths[index] = delta_ij_u[0]
    index += 1

    tree[index] = np.array((true_indices[2], intermediate_node))
    branch_lengths[index] = delta_ij_u[1]
    index += 1

    tree[index] = np.array((true_indices[0], intermediate_node))
    branch_lengths[index] = 0.5 * (
        distance_matrix[1, 0] + distance_matrix[2, 0] - distance_matrix[1, 2]
    )
    index += 1

    return tree[:index], branch_lengths[:index]


# Q matrix calculation + minimum i, j (step 1 & 2)
@nb.njit
# @numba_cc.export('_find_join_nodes', '(f64[:])')
def _find_join_nodes(distance_matrix):
    """
    Finds which nodes to join next

    Parameters
    ----------
    distance_matrix

    Returns
    -------
    indices (i, j) of nodes to join
    """
    n = distance_matrix.shape[0]
    q_matrix = np.zeros((n, n))
    q_matrix[:] = np.inf
    min_ij = np.array([0, 0], dtype=np.uint64)
    min_q = np.inf
    for i in range(n):
        for j in range(n):
            if i != j:
                q_matrix[i, j] = (
                    (n - 2) * distance_matrix[i, j]
                    - np.sum(distance_matrix[i, :])
                    - np.sum(distance_matrix[j, :])
                )
                # q_matrix[i, j] = distance_matrix[i, j]
                if q_matrix[i, j] < min_q:
                    min_ij[0], min_ij[1] = i, j
                    min_q = q_matrix[i, j]
    return min_ij


# Branch length calculation (step 3)
@nb.njit
# @numba_cc.export('_find_branch_length', '(f64[:], i64, i64)')
def _find_branch_length(distance_matrix, i, j):
    """
    Finds branch lengths of old nodes to newly created node

    Parameters
    ----------
    distance_matrix
    i
        first node to join
    j
        second node to join

    Returns
    -------
    (branch length of i to new node, branch length of j to new node)
    """
    n = distance_matrix.shape[0]
    delta_i_u = 0.5 * distance_matrix[i, j] + (0.5 / (n - 2)) * (
        np.sum(distance_matrix[i, :]) - np.sum(distance_matrix[j, :])
    )
    delta_j_u = distance_matrix[i, j] - delta_i_u
    return np.array([delta_i_u, delta_j_u])



import numba as nb
import numpy as np
from geometricus import MomentInvariants, SplitType, GeometricusEmbedding, MomentType
from caretta import dynamic_time_warping as dtw, score_functions, helper

"""
Provides pairwise superposition functions to use in Caretta
Each takes coords_1, coords_2, parameters as input
    parameters is a dict with gap_open_penalty, gap_extend_penalty, gamma, and other function-specific parameters as keys

returns score, superposed_coords_1, superposed_coords_2
"""


def dtw_svd_superpose_function(
    coords_1, coords_2, parameters: dict,
):
    """
    Assumes coords_1 and coords_2 are already in a well-superposed state,
    runs DTW alignment and then superposes with Kabsch on the aligning positions
    """
    score_matrix = score_functions.make_score_matrix(
        coords_1,
        coords_2,
        score_functions.get_caretta_score,
        parameters["gamma"],
        normalized=False,
    )
    _, coords_1, coords_2, common_coords_1, common_coords_2 = _align_and_superpose(
        coords_1,
        coords_2,
        score_matrix,
        parameters["gap_open_penalty"],
        parameters["gap_extend_penalty"],
    )
    return (
        score_functions.get_total_score(
            common_coords_1,
            common_coords_2,
            score_functions.get_caretta_score,
            parameters["gamma"],
            False,
        ),
        coords_1,
        coords_2,
    )


def signal_superpose_function(coords_1, coords_2, parameters):
    """
    Makes initial superposition of coordinates using DTW alignment of overlapping signals
    A signal is a vector of euclidean distances of first (or last) coordinate to all others in a "size"-residue stretch
    """
    score_first, c1_first, c2_first = _signal_superpose_index(
        0,
        coords_1,
        coords_2,
        parameters["gap_open_penalty"],
        parameters["gap_extend_penalty"],
        size=parameters["size"],
    )
    score_last, c1_last, c2_last = _signal_superpose_index(
        -1,
        coords_1,
        coords_2,
        parameters["gap_open_penalty"],
        parameters["gap_extend_penalty"],
        size=parameters["size"],
    )
    if score_first > score_last:
        return score_first, c1_first, c2_first
    else:
        return score_last, c1_last, c2_last


def signal_svd_superpose_function(coords_1, coords_2, parameters):
    """
    Uses signal_superpose followed by dtw_svd_superpose
    """
    _, coords_1, coords_2 = signal_superpose_function(coords_1, coords_2, parameters)
    return dtw_svd_superpose_function(coords_1, coords_2, parameters)


def moment_superpose_function(coords_1, coords_2, parameters):
    """
    Uses 4 rotation/translation invariant moments for each "split_size"-mer to run DTW
    """
    if "upsample_rate" not in parameters:
        parameters["upsample_rate"] = 10
    if "moment_types" not in parameters:
        parameters["moment_types"] = ["O_3", "O_4", "O_5", "F"]
    if "scale" not in parameters:
        parameters["scale"] = True
    if "gamma_moment" not in parameters:
        parameters["gamma_moment"] = 0.6
    if "gamma" not in parameters:
        parameters["gamma"] = 0.03
    if "gap_open_penalty" not in parameters:
        parameters["gap_open_penalty"] = 0.0
    if "gap_extend_penalty" not in parameters:
        parameters["gap_extend_penalty"] = 0.0

    moment_types = [MomentType[x] for x in parameters["moment_types"]]
    moments_1 = MomentInvariants.from_coordinates(
        "name",
        coords_1,
        split_type=SplitType[parameters["split_type"]],
        split_size=parameters["split_size"],
        upsample_rate=parameters["upsample_rate"],
        moment_types=moment_types,
    ).moments
    moments_2 = MomentInvariants.from_coordinates(
        "name",
        coords_2,
        split_type=SplitType[parameters["split_type"]],
        split_size=parameters["split_size"],
        upsample_rate=parameters["upsample_rate"],
        moment_types=moment_types,
    ).moments
    if parameters["scale"]:
        moments_1 = np.log1p(moments_1)
        moments_2 = np.log1p(moments_2)
    score_matrix = score_functions.make_score_matrix(
        moments_1,
        moments_2,
        score_functions.get_caretta_score,
        parameters["gamma_moment"],
        normalized=True,
    )
    score, coords_1, coords_2, _, _ = _align_and_superpose(
        coords_1,
        coords_2,
        score_matrix,
        parameters["gap_open_penalty"],
        parameters["gap_extend_penalty"],
    )
    return score, coords_1, coords_2


def geometricus_superpose_function(coords_1, coords_2, parameters):
    if "upsample_rate" not in parameters:
        parameters["upsample_rate"] = 10
    invariants = [
        MomentInvariants.from_coordinates(
            "name1",
            coords_1,
            split_type=SplitType[parameters[f"split_type"]],
            split_size=parameters[f"split_size"],
            upsample_rate=parameters["upsample_rate"],
        ),
        MomentInvariants.from_coordinates(
            "name2",
            coords_2,
            split_type=SplitType[parameters[f"split_type"]],
            split_size=parameters[f"split_size"],
            upsample_rate=parameters["upsample_rate"],
        ),
    ]
    embedder = GeometricusEmbedding.from_invariants(
        invariants, resolution=parameters["resolution"], protein_keys=["name1", "name2"]
    )
    score_matrix = score_functions.make_score_matrix(
        np.array(embedder.proteins_to_shapemers["name1"]),
        np.array(embedder.proteins_to_shapemers["name2"]),
        score_functions.get_caretta_score,
        gamma=parameters["gamma"],
        normalized=False,
    )
    score, coords_1, coords_2, _, _ = _align_and_superpose(
        coords_1,
        coords_2,
        score_matrix,
        parameters["gap_open_penalty"],
        parameters["gap_extend_penalty"],
    )
    return score, coords_1, coords_2


def geometricus_svd_superpose_function(coords_1, coords_2, parameters):
    """
    Uses moment_multiple_superpose followed by dtw_svd_superpose
    """
    _, coords_1, coords_2 = geometricus_superpose_function(
        coords_1, coords_2, parameters
    )
    return dtw_svd_superpose_function(coords_1, coords_2, parameters)


def moment_multiple_superpose_function(coords_1, coords_2, parameters):
    """
    Uses 4 rotation/translation invariant moments for each "split_size"-mer with different fragmentation approaches to run DTW
    """
    moments_1 = []
    moments_2 = []
    if "upsample_rate" not in parameters:
        parameters["upsample_rate"] = 10
    for i in range(parameters["num_split_types"]):
        if f"moment_types_{i}" not in parameters:
            parameters[f"moment_types_{i}"] = ["O_3", "O_4", "O_5", "F"]
    if "scale" not in parameters:
        parameters["scale"] = True
    if "gamma_moment" not in parameters:
        parameters["gamma_moment"] = 0.6
    if "gamma" not in parameters:
        parameters["gamma"] = 0.03
    if "gap_open_penalty" not in parameters:
        parameters["gap_open_penalty"] = 0.0
    if "gap_extend_penalty" not in parameters:
        parameters["gap_extend_penalty"] = 0.0

    for i in range(parameters["num_split_types"]):
        moment_types = [MomentType[x] for x in parameters[f"moment_types_{i}"]]
        moments_1_1 = MomentInvariants.from_coordinates(
            "name",
            coords_1,
            split_type=SplitType[parameters[f"split_type_{i}"]],
            split_size=parameters[f"split_size_{i}"],
            upsample_rate=parameters["upsample_rate"],
            moment_types=moment_types,
        ).moments
        moments_2_1 = MomentInvariants.from_coordinates(
            "name",
            coords_2,
            split_type=SplitType[parameters[f"split_type_{i}"]],
            split_size=parameters[f"split_size_{i}"],
            upsample_rate=parameters["upsample_rate"],
            moment_types=moment_types,
        ).moments
        if parameters["scale"]:
            moments_1_1 = np.log1p(moments_1_1)
            moments_2_1 = np.log1p(moments_2_1)
        moments_1.append(moments_1_1)
        moments_2.append(moments_2_1)
    score_matrix = np.zeros((moments_1[0].shape[0], moments_2[0].shape[0]))
    for (m_1, m_2) in zip(moments_1, moments_2):
        score_matrix += score_functions.make_score_matrix(
            m_1,
            m_2,
            score_functions.get_caretta_score,
            gamma=parameters["gamma_moment"],
            normalized=True,
        )
    score, coords_1, coords_2, _, _ = _align_and_superpose(
        coords_1,
        coords_2,
        score_matrix,
        parameters["gap_open_penalty"],
        parameters["gap_extend_penalty"],
    )
    return score, coords_1, coords_2


def moment_multiple_svd_superpose_function(coords_1, coords_2, parameters):
    """
    Uses moment_multiple_superpose followed by dtw_svd_superpose
    """
    _, coords_1, coords_2 = moment_multiple_superpose_function(
        coords_1, coords_2, parameters
    )
    return dtw_svd_superpose_function(coords_1, coords_2, parameters)


def moment_svd_superpose_function(coords_1, coords_2, parameters):
    """
    Uses moment_superpose followed by dtw_svd_superpose
    """
    _, coords_1, coords_2 = moment_superpose_function(coords_1, coords_2, parameters)
    return dtw_svd_superpose_function(coords_1, coords_2, parameters)


@nb.njit
def _align_and_superpose(
    coords_1, coords_2, score_matrix, gap_open_penalty, gap_extend_penalty
):
    """
    Runs DTW on a score matrix and Kabsch superposition on resulting alignment
    """
    dtw_aln_array_1, dtw_aln_array_2, score = dtw.dtw_align(
        score_matrix, gap_open_penalty, gap_extend_penalty
    )
    pos_1, pos_2 = helper.get_common_positions(dtw_aln_array_1, dtw_aln_array_2)
    common_coords_1, common_coords_2 = coords_1[pos_1], coords_2[pos_2]
    coords_1, coords_2, common_coords_2 = paired_svd_superpose_with_subset(
        coords_1, coords_2, common_coords_1, common_coords_2
    )
    return score, coords_1, coords_2, common_coords_1, common_coords_2


@nb.njit
def paired_svd_superpose(coords_1: np.ndarray, coords_2: np.ndarray):
    """
    Superpose paired coordinates on each other using Kabsch superposition (SVD)

    Parameters
    ----------
    coords_1
        numpy array of coordinate data for the first protein; shape = (n, 3)
    coords_2
        numpy array of corresponding coordinate data for the second protein; shape = (n, 3)

    Returns
    -------
    rotation matrix, translation matrix for optimal superposition
    """
    centroid_1, centroid_2 = (
        helper.nb_mean_axis_0(coords_1),
        helper.nb_mean_axis_0(coords_2),
    )
    coords_1_c, coords_2_c = coords_1 - centroid_1, coords_2 - centroid_2
    correlation_matrix = np.dot(coords_2_c.T, coords_1_c)
    u, s, v = np.linalg.svd(correlation_matrix)
    reflect = np.linalg.det(u) * np.linalg.det(v) < 0
    if reflect:
        s[-1] = -s[-1]
        u[:, -1] = -u[:, -1]
    rotation_matrix = np.dot(u, v)
    translation_matrix = centroid_1 - np.dot(centroid_2, rotation_matrix)
    return rotation_matrix.astype(np.float64), translation_matrix.astype(np.float64)


@nb.njit
def paired_svd_superpose_with_subset(
    coords_1, coords_2, common_coords_1, common_coords_2
):
    """
    Superpose two sets of un-aligned coordinates using smaller subsets of aligned coordinates

    Parameters
    ----------
    coords_1
    coords_2
    common_coords_1
    common_coords_2

    Returns
    -------
    superposed coord_1, superposed coords_2, superposed common_coords_2
    """
    rot, tran = paired_svd_superpose(common_coords_1, common_coords_2)
    coords_1 = coords_1 - helper.nb_mean_axis_0(common_coords_1)
    coords_2 = np.dot(coords_2 - helper.nb_mean_axis_0(common_coords_2), rot)
    common_coords_2_rot = apply_rotran(common_coords_2, rot, tran)
    return coords_1, coords_2, common_coords_2_rot


def _signal_superpose_index(
    index,
    coords_1,
    coords_2,
    gap_open_penalty=0.0,
    gap_extend_penalty=0.0,
    size=30,
    overlap=1,
):
    """
    Makes initial superposition using DTW alignment of overlapping signals
    A signal is a vector of euclidean distances of first (or last) coordinate to all others in a 30-residue stretch
    """

    def _make_signal_index(coords, idx):
        centroid = coords[idx]
        distances = np.zeros(coords.shape[0])
        for c in range(coords.shape[0]):
            distances[c] = np.sqrt(np.sum((coords[c] - centroid) ** 2, axis=-1))
        return distances

    signals_1 = np.zeros(((coords_1.shape[0] - size) // overlap, size))
    signals_2 = np.zeros(((coords_2.shape[0] - size) // overlap, size))
    middles_1 = np.zeros((signals_1.shape[0], coords_1.shape[1]))
    middles_2 = np.zeros((signals_2.shape[0], coords_2.shape[1]))
    if index == -1:
        index = size - 1
    for x, i in enumerate(range(0, signals_1.shape[0] * overlap, overlap)):
        signals_1[x] = _make_signal_index(coords_1[i : i + size], index)
        middles_1[x] = coords_1[i + index]
    for x, i in enumerate(range(0, signals_2.shape[0] * overlap, overlap)):
        signals_2[x] = _make_signal_index(coords_2[i : i + size], index)
        middles_2[x] = coords_2[i + index]
    score_matrix = score_functions.make_score_matrix(
        signals_1,
        signals_2,
        score_functions.get_signal_score,
        gamma=0.1,
        normalized=False,
    )
    dtw_1, dtw_2, score = dtw.dtw_align(
        score_matrix, gap_open_penalty, gap_extend_penalty
    )
    pos_1, pos_2 = helper.get_common_positions(dtw_1, dtw_2)
    aln_coords_1 = np.zeros((len(pos_1), coords_1.shape[1]))
    aln_coords_2 = np.zeros((len(pos_2), coords_2.shape[1]))
    for i, (p1, p2) in enumerate(zip(pos_1, pos_2)):
        aln_coords_1[i] = middles_1[p1]
        aln_coords_2[i] = middles_2[p2]
    coords_1, coords_2, _ = paired_svd_superpose_with_subset(
        coords_1, coords_2, aln_coords_1, aln_coords_2
    )
    return score, coords_1, coords_2


@nb.njit
def svd_superimpose(coords_1: np.ndarray, coords_2: np.ndarray):
    """
    Superimpose paired coordinates on each other using svd

    Parameters
    ----------
    coords_1
        numpy array of coordinate data for the first protein; shape = (n, 3)
    coords_2
        numpy array of corresponding coordinate data for the second protein; shape = (n, 3)

    Returns
    -------
    rotation matrix, translation matrix for optimal superposition
    """
    centroid_1, centroid_2 = (
        helper.nb_mean_axis_0(coords_1),
        helper.nb_mean_axis_0(coords_2),
    )
    coords_1_c, coords_2_c = coords_1 - centroid_1, coords_2 - centroid_2
    correlation_matrix = np.dot(coords_2_c.T, coords_1_c)
    u, s, v = np.linalg.svd(correlation_matrix)
    reflect = np.linalg.det(u) * np.linalg.det(v) < 0
    if reflect:
        s[-1] = -s[-1]
        u[:, -1] = -u[:, -1]
    rotation_matrix = np.dot(u, v)
    translation_matrix = centroid_1 - np.dot(centroid_2, rotation_matrix)
    return rotation_matrix.astype(np.float64), translation_matrix.astype(np.float64)


@nb.njit
def apply_rotran(
    coords: np.ndarray, rotation_matrix: np.ndarray, translation_matrix: np.ndarray
) -> np.ndarray:
    """
    Applies a rotation and translation matrix onto coordinates

    Parameters
    ----------
    coords
    rotation_matrix
    translation_matrix

    Returns
    -------
    transformed coordinates
    """
    return np.dot(coords, rotation_matrix) + translation_matrix


@nb.njit
def superpose_with_pos(coords_1, coords_2, common_coords_1, common_coords_2):
    """
    Superpose two sets of un-aligned coordinates using smaller subsets of aligned coordinates

    Parameters
    ----------
    coords_1
    coords_2
    common_coords_1
    common_coords_2

    Returns
    -------
    superposed coord_1, superposed coords_2, superposed common_coords_2
    """
    rot, tran = svd_superimpose(common_coords_1, common_coords_2)
    coords_1 = coords_1 - helper.nb_mean_axis_0(common_coords_1)
    coords_2 = np.dot(coords_2 - helper.nb_mean_axis_0(common_coords_2), rot)
    common_coords_2_rot = apply_rotran(common_coords_2, rot, tran)
    return coords_1, coords_2, common_coords_2_rot



import subprocess
import typing
from typing import List, Union, Tuple
from pathlib import Path, PosixPath
import Bio.PDB
import numba as nb
import numpy as np
import prody as pd


def secondary_to_array(secondary):
    return np.array(secondary, dtype="S1").view(np.int8)


@nb.njit
def get_common_positions(aln_array_1, aln_array_2):
    """
    Return positions where neither alignment has a gap (-1)

    Parameters
    ----------
    aln_array_1
    aln_array_2

    Returns
    -------
    common_positions_1, common_positions_2
    """
    pos_1 = np.array(
        [
            aln_array_1[i]
            for i in range(len(aln_array_1))
            if aln_array_1[i] != -1 and aln_array_2[i] != -1
        ],
        dtype=np.int64,
    )
    pos_2 = np.array(
        [
            aln_array_2[i]
            for i in range(len(aln_array_2))
            if aln_array_1[i] != -1 and aln_array_2[i] != -1
        ],
        dtype=np.int64,
    )
    return pos_1, pos_2


@nb.njit
def nb_mean_axis_0(array: np.ndarray) -> np.ndarray:
    """
    Same as np.mean(array, axis=0) but njitted
    """
    mean_array = np.zeros(array.shape[1])
    for i in range(array.shape[1]):
        mean_array[i] = np.mean(array[:, i])
    return mean_array


@nb.njit
def nb_std_axis_0(array: np.ndarray) -> np.ndarray:
    """
    Same as np.std(array, axis=0) but njitted
    """
    std_array = np.zeros(array.shape[1])
    for i in range(array.shape[1]):
        std_array[i] = np.std(array[:, i])
    return std_array


@nb.njit
def normalize(numbers):
    minv, maxv = np.min(numbers), np.max(numbers)
    return (numbers - minv) / (maxv - minv)


def get_file_parts(input_filename: Union[str, Path]) -> Tuple[str, str, str]:
    """
    Gets directory path, name, and extension from a filename
    Parameters
    ----------
    input_filename

    Returns
    -------
    (path, name, extension)
    """
    input_filename = Path(input_filename)
    path = str(input_filename.parent)
    extension = input_filename.suffix
    name = input_filename.stem
    return path, name, extension


def get_alpha_indices(protein: pd.AtomGroup) -> List[int]:
    """
    Get indices of alpha carbons of pd AtomGroup object
    """
    return [a.getIndex() for a in protein.iterAtoms() if a.getName() == "CA"]


def get_beta_indices(protein: pd.AtomGroup) -> List[int]:
    """
    Get indices of beta carbons of pd AtomGroup object
    (If beta carbon doesn't exist, alpha carbon index is returned)
    """
    residue_splits = group_indices(protein.getResindices())
    i = 0
    indices = []
    for split in residue_splits:
        ca = None
        cb = None
        for _ in split:
            if protein[i].getName() == "CB":
                cb = protein[i].getIndex()
            if protein[i].getName() == "CA":
                ca = protein[i].getIndex()
            i += 1
        if cb is not None:
            indices.append(cb)
        else:
            assert ca is not None
            indices.append(ca)
    return indices


def group_indices(input_list: List[int]) -> List[List[int]]:
    """
    [1, 1, 1, 2, 2, 3, 3, 3, 4] -> [[0, 1, 2], [3, 4], [5, 6, 7], [8]]
    Parameters
    ----------
    input_list

    Returns
    -------
    list of lists
    """
    output_list = []
    current_list = []
    current_index = None
    for i in range(len(input_list)):
        if current_index is None:
            current_index = input_list[i]
        if input_list[i] == current_index:
            current_list.append(i)
        else:
            output_list.append(current_list)
            current_list = [i]
        current_index = input_list[i]
    output_list.append(current_list)
    return output_list


def clustal_msa_from_sequences(
    sequence_file, alignment_file, hmm_file=None, distance_matrix_file=None
):
    """
    Align sequences optionally using hmm_file as a guide

    Parameters
    ----------
    sequence_file
    alignment_file
    hmm_file
    distance_matrix_file
        Writes pairwise distance matrix into file if not None

    Returns
    -------

    """
    if hmm_file is not None:
        if distance_matrix_file is None:
            subprocess.check_call(
                f"clustalo "
                f"--output-order=input-order "
                f"--log=log_clustal.out "
                f"-i {sequence_file} "
                f"--hmm-in={hmm_file} "
                f"-o {alignment_file} "
                f"--threads=10 --force -v",
                shell=True,
            )
        else:
            subprocess.check_call(
                f"clustalo "
                f"--output-order=input-order "
                f"--log=log_clustal.out --full "
                f"--distmat-out={distance_matrix_file} "
                f"-i {sequence_file} "
                f"--hmm-in={hmm_file} "
                f"-o {alignment_file} "
                f"--threads=10 --force -v",
                shell=True,
            )
    else:
        if distance_matrix_file is None:
            subprocess.check_call(
                f"clustalo "
                f"--output-order=input-order "
                f"--log=log_clustal.out "
                f"-i {sequence_file} "
                f"-o {alignment_file} "
                f"--threads=10 --force -v",
                shell=True,
            )
        else:
            subprocess.check_call(
                f"clustalo "
                f"--output-order=input-order "
                f"--log=log_clustal.out "
                f"--full "
                f"--distmat-out={distance_matrix_file} "
                f"-i {sequence_file} "
                f"-o {alignment_file} "
                f"--threads=10 --force -v",
                shell=True,
            )


def get_sequences_from_fasta(
    fasta_file: Union[str, Path], prune_headers: bool = True
) -> dict:
    """
    Returns dict of accession to sequence from fasta file
    Parameters
    ----------
    fasta_file
    prune_headers
        only keeps accession upto first /

    Returns
    -------
    {accession:sequence}
    """
    sequences = {}
    with open(fasta_file) as f:
        current_sequence = []
        current_key = None
        for line in f:
            if not len(line.strip()):
                continue
            if line.startswith(">"):
                if current_key is None:
                    if "/" in line and prune_headers:
                        current_key = line.split(">")[1].split("/")[0].strip()
                    else:
                        current_key = line.split(">")[1].strip()
                else:
                    sequences[current_key] = "".join(current_sequence)
                    current_sequence = []
                    if "/" in line and prune_headers:
                        current_key = line.split(">")[1].split("/")[0].strip()
                    else:
                        current_key = line.split(">")[1].strip()
            else:
                current_sequence.append(line.strip())
        sequences[current_key] = "".join(current_sequence)
    return sequences


def read_pdb(input_file, name: str = None, chain: str = None) -> tuple:
    """
    returns protein information from PDB file
    Parameters
    ----------
    input_file
    name
        None => takes from filename
    chain
        only for that chain

    Returns
    -------
    structure Object, list of residue Objects, list of peptide Objects, sequence, sequence to residue index
    """
    if name is None:
        (_, name, _) = get_file_parts(input_file)
    input_file = str(input_file)
    structure = Bio.PDB.PDBParser().get_structure(name, input_file)
    if chain is not None:
        structure = structure[0][chain]
    residues = Bio.PDB.Selection.unfold_entities(structure, "R")
    peptides = Bio.PDB.PPBuilder().build_peptides(structure)
    sequence = "".join([str(peptide.get_sequence()) for peptide in peptides])
    residue_dict = dict(zip(residues, range(len(residues))))
    seq_to_res_index = [residue_dict[r] for peptide in peptides for r in peptide]
    return structure, residues, peptides, sequence, seq_to_res_index


def get_alpha_coordinates(residue) -> np.ndarray:
    """
    Returns alpha coordinates of BioPython residue object
    """
    return np.array(residue["CA"].get_coord())


def get_beta_coordinates(residue) -> np.ndarray:
    """
    Returns beta coordinates of BioPython residue object
    (alpha if Gly)
    """
    if residue.get_resname() == "GLY" or "CB" not in residue:
        return get_alpha_coordinates(residue)
    return np.array(residue["CB"].get_coord())


def parse_pdb_files(input_pdb):
    if type(input_pdb) == str or type(input_pdb) == PosixPath:
        input_pdb = Path(input_pdb)
        if input_pdb.is_dir():
            pdb_files = list(input_pdb.glob("*.pdb"))
        elif input_pdb.is_file():
            with open(input_pdb) as f:
                pdb_files = f.read().strip().split("\n")
        else:
            pdb_files = str(input_pdb).split("\n")
    else:
        pdb_files = list(input_pdb)
        if not Path(pdb_files[0]).is_file():
            pdb_files = [pd.fetchPDB(pdb_name) for pdb_name in pdb_files]
    return pdb_files


def parse_pdb_files_and_clean(
    input_pdb: str, output_pdb: Union[str, Path] = "./cleaned_pdb",
) -> List[Union[str, Path]]:
    if not Path(output_pdb).exists():
        Path(output_pdb).mkdir()
    pdb_files = parse_pdb_files(input_pdb)
    output_pdb_files = []
    for pdb_file in pdb_files:
        pdb = pd.parsePDB(pdb_file).select("protein")
        chains = pdb.getChids()
        if len(chains) and len(chains[0].strip()):
            pdb = pdb.select(f"chain {chains[0]}")
        output_pdb_file = str(Path(output_pdb) / f"{Path(pdb_file).stem}.pdb")
        pd.writePDB(output_pdb_file, pdb)
        output_pdb_files.append(output_pdb_file)
    return output_pdb_files


def write_distance_matrix(
    names: typing.List[str],
    distance_matrix: np.ndarray,
    filename: typing.Union[Path, str],
):
    """
    Writes distance matrix to file in clustal format

    Parameters
    ----------
    names
        protein names in order
    distance_matrix
    filename
    """
    with open(filename, "w") as f:
        f.write(f"{len(names)}\n")
        for i in range(len(names)):
            row = " ".join(f"{x:.4f}" for x in distance_matrix[i])
            f.write(f"{names[i]} {row}\n")


def read_distance_matrix(filename: typing.Union[Path, str]):
    """
    Read distance matrix file

    Parameters
    ----------
    filename

    Returns
    -------
    protein names in order, distance matrix
    """
    names = []
    num_proteins = 0
    with open(filename) as f:
        for i, line in enumerate(f):
            if i == 0:
                num_proteins = int(line.strip())
                continue
            names.append(line.split()[0].strip().split("/")[0].strip())
    assert len(names) == num_proteins
    distance_matrix = np.loadtxt(
        filename, skiprows=1, usecols=range(1, num_proteins + 1)
    )
    return names, distance_matrix



import multiprocessing
from pathlib import Path

import numpy as np
import prody as pd
import Bio.PDB
from Bio.PDB.ResidueDepth import get_surface, residue_depth, ca_depth, min_dist
from caretta import helper


def read_pdb(input_file, name: str = None, chain: str = None) -> tuple:
    """
    returns protein information from PDB file
    Parameters
    ----------
    input_file
    name
        None => takes from filename
    chain
        only for that chain

    Returns
    -------
    structure Object, list of residue Objects, list of peptide Objects, sequence, sequence to residue index
    """
    if name is None:
        name = Path(input_file).stem
    input_file = str(input_file)
    structure = Bio.PDB.PDBParser().get_structure(name, input_file)
    if chain is not None:
        structure = structure[0][chain]
    residues = Bio.PDB.Selection.unfold_entities(structure, "R")
    peptides = Bio.PDB.PPBuilder().build_peptides(structure)
    sequence = "".join([str(peptide.get_sequence()) for peptide in peptides])
    residue_dict = dict(zip(residues, range(len(residues))))
    seq_to_res_index = [residue_dict[r] for peptide in peptides for r in peptide]
    return structure, residues, peptides, sequence, seq_to_res_index


def get_alpha_coordinates(residue) -> np.ndarray:
    """
    Returns alpha coordinates of BioPython residue object
    """
    return np.array(residue["CA"].get_coord())


def get_beta_coordinates(residue) -> np.ndarray:
    """
    Returns beta coordinates of BioPython residue object
    (alpha if Gly)
    """
    if residue.get_resname() == "GLY" or "CB" not in residue:
        return get_alpha_coordinates(residue)
    return np.array(residue["CB"].get_coord())


def get_residue_depths(pdb_file):
    """
    Get residue depths

    Parameters
    ----------
    pdb_file

    Returns
    -------
    dict of depth _ ca/cb/mean
    """
    structure, residues, _, _, _ = read_pdb(pdb_file)
    surface = get_surface(structure)
    data = {
        "depth_mean": np.array(
            [residue_depth(residue, surface) for residue in residues]
        ),
        "depth_cb": np.array(
            [min_dist(get_beta_coordinates(residue), surface) for residue in residues]
        ),
        "depth_ca": np.array([ca_depth(residue, surface) for residue in residues]),
    }
    return data


def get_fluctuations(protein: pd.AtomGroup, n_modes: int = 50):
    """
    Get atom fluctuations using anisotropic and Gaussian network models with n_modes modes.

    Parameters
    ----------
    protein
    n_modes

    Returns
    -------
    dict of anm_ca, anm_cb, gnm_ca, gnm_cb
    """
    data = {}
    beta_indices = helper.get_beta_indices(protein)
    alpha_indices = helper.get_alpha_indices(protein)
    data["anm_cb"] = get_anm_fluctuations(protein[beta_indices], n_modes)
    data["gnm_cb"] = get_gnm_fluctuations(protein[beta_indices], n_modes)
    data["anm_ca"] = get_anm_fluctuations(protein[alpha_indices], n_modes)
    data["gnm_ca"] = get_gnm_fluctuations(protein[alpha_indices], n_modes)
    return data


def get_anm_fluctuations(protein: pd.AtomGroup, n_modes: int = 50):
    """
    Get atom fluctuations using an Anisotropic network model with n_modes modes.
    """
    protein_anm, _ = pd.calcANM(protein, n_modes=n_modes, selstr="all")
    return pd.calcSqFlucts(protein_anm)


def get_gnm_fluctuations(protein: pd.AtomGroup, n_modes: int = 50):
    """
    Get atom fluctuations using a Gaussian network model with n_modes modes.
    """
    protein_gnm, _ = pd.calcGNM(protein, n_modes=n_modes, selstr="all")
    return pd.calcSqFlucts(protein_gnm)


def get_features_multiple(
    pdb_files, dssp_dir, num_threads=20, only_dssp=True, force_overwrite=True, n_modes=50
):
    """
    Extract features for a list of pdb_files in parallel

    Parameters
    ----------
    pdb_files
    dssp_dir
        directory to store tmp dssp files
    num_threads
    only_dssp
        extract only dssp features
    force_overwrite
        force rerun DSSP

    Returns
    -------
    List of feature dicts (same order as pdb_files)
    """
    num_threads = min(len(pdb_files), num_threads)
    with multiprocessing.Pool(processes=num_threads) as pool:
        return pool.starmap(
            get_features,
            [
                (pdb_file, dssp_dir, only_dssp, force_overwrite, n_modes)
                for pdb_file in pdb_files
            ],
        )


def get_features(pdb_file: str, dssp_dir: str, only_dssp=True, force_overwrite=True, n_modes=50):
    """
    Extract features from a pdb_file

    Parameters
    ----------
    pdb_file
    dssp_dir
        directory to store tmp dssp files
    only_dssp
        extract only dssp features (use if not interested in features)
    force_overwrite
        force rerun DSSP

    Returns
    -------
    dict of features
    """
    pdb_file = str(pdb_file)
    name = Path(pdb_file).stem
    protein = pd.parsePDB(pdb_file).select("protein")
    pdb_file = str(Path(dssp_dir) / f"{name}.pdb")
    pd.writePDB(pdb_file, protein)
    protein = pd.parsePDB(pdb_file)
    dssp_file = Path(dssp_dir) / f"{name}.dssp"
    if force_overwrite or not dssp_file.exists():
        dssp_file = pd.execDSSP(str(pdb_file), outputname=name, outputdir=str(dssp_dir))
    protein = pd.parseDSSP(dssp=str(dssp_file), ag=protein, parseall=True)
    data = get_dssp_features(protein)
    if only_dssp:
        return data
    else:
        data = {**data, **get_fluctuations(protein, n_modes)}
        try:
            data = {**data, **get_residue_depths(pdb_file)}
        except RuntimeError as e:
            print(f"Failed to calculate residue depths: {e}")
        return data


def get_dssp_features(protein_dssp):
    """
    Extracts DSSP features (assumes DSSP is run already)

    Parameters
    ----------
    protein_dssp
        protein on which execDSSP has been called
    Returns
    -------
    dict of secondary,
    dssp_
        NH_O_1_index, NH_O_1_energy
            hydrogen bonds; e.g. -3,-1.4 means: if this residue is residue i then N-H of I is h-bonded to C=O of I-3 with an
            electrostatic H-bond energy of -1.4 kcal/mol. There are two columns for each type of H-bond, to allow for bifurcated H-bonds.
        NH_O_2_index, NH_O_2_energy
        O_NH_1_index, O_NH_1_energy
        O_NH_2_index, O_NH_2_energy
        acc
            number of water molecules in contact with this residue *10. or residue water exposed surface in Angstrom^2.
        alpha
            virtual torsion angle (dihedral angle) defined by the four Cα atoms of residues I-1,I,I+1,I+2. Used to define chirality.
        kappa
            virtual bond angle (bend angle) defined by the three Cα atoms of residues I-2,I,I+2. Used to define bend (structure code ‘S’).
        phi
            IUPAC peptide backbone torsion angles.
        psi
            IUPAC peptide backbone torsion angles.
        tco
            cosine of angle between C=O of residue I and C=O of residue I-1. For α-helices, TCO is near +1, for β-sheets TCO is near -1.
    Ignores:
    dssp_bp1, dssp_bp2, and dssp_sheet_label: residue number of first and second bridge partner followed by one letter sheet label
    """
    dssp_ignore = ["dssp_bp1", "dssp_bp2", "dssp_sheet_label", "dssp_resnum"]
    dssp_labels = [
        label
        for label in protein_dssp.getDataLabels()
        if label.startswith("dssp") and label not in dssp_ignore
    ]
    data = {}
    alpha_indices = helper.get_alpha_indices(protein_dssp)
    indices = [protein_dssp[x].getData("dssp_resnum") for x in alpha_indices]
    assert len(alpha_indices) == len(indices)
    for label in dssp_labels + ["secondary"]:
        label_to_index = {
            i - 1: protein_dssp[x].getData(label)
            for i, x in zip(indices, alpha_indices)
        }
        data[label] = np.array(
            [
                label_to_index[i] if i in label_to_index else 0
                for i in range(len(alpha_indices))
            ]
        )
    return data



import base64
import datetime
import pickle
from dataclasses import dataclass
from pathlib import Path
import numpy as np
import prody as pd
import requests as rq

import typing

from caretta import multiple_alignment


def heatmap(data):
    return dict(
        data=[dict(z=data, type="heatmap", showscale=False)],
        layout=dict(margin=dict(l=25, r=25, t=25, b=25)),
    )


def empty_dict():
    data = [dict(z=np.zeros((2, 2)), type="heatmap", showscale=False)]
    layout = dict(margin=dict(l=25, r=25, t=25, b=25))
    return dict(data=data, layout=layout)


def empty_object(suite):
    return compress_object(np.zeros(0), suite)


def get_estimated_time(msa_class: multiple_alignment.StructureMultiple):
    n = len(msa_class.structures)
    l = max(s.length for s in msa_class.structures)
    func = lambda x, r: (x[0] ** 2 * r * x[1] ** 2)
    return str(datetime.timedelta(seconds=int(func((l, n), 9.14726052e-06))))


def line(data):
    y = np.array([np.nanmean(data[:, x]) for x in range(data.shape[1])])
    y_se = np.array(
        [np.nanstd(data[:, x]) / np.sqrt(data.shape[1]) for x in range(data.shape[1])]
    )

    data = [
        dict(
            y=list(y + y_se) + list(y - y_se)[::-1],
            x=list(range(data.shape[1])) + list(range(data.shape[1]))[::-1],
            fillcolor="lightblue",
            fill="toself",
            type="scatter",
            mode="lines",
            name="Standard error",
            line=dict(color="lightblue"),
        ),
        dict(
            y=y,
            x=np.arange(data.shape[1]),
            type="scatter",
            mode="lines",
            name="Mean",
            line=dict(color="blue"),
        ),
    ]
    return dict(
        data=data,
        layout=dict(legend=dict(x=0.5, y=1.2), margin=dict(l=25, r=25, t=25, b=25)),
    )


def scatter3D(coordinates_dict):
    data = []
    for k, v in coordinates_dict.items():
        x, y, z = v[:, 0], v[:, 1], v[:, 2]
        data.append(
            dict(
                x=x,
                y=y,
                z=z,
                mode="lines",
                type="scatter3d",
                text=None,
                name=str(k),
                line=dict(width=3, opacity=0.8),
            )
        )
    layout = dict(
        margin=dict(l=20, r=20, t=20, b=20),
        clickmode="event+select",
        scene=dict(
            xaxis=dict(visible=False, showgrid=False, showline=False),
            yaxis=dict(visible=False, showgrid=False, showline=False),
            zaxis=dict(visible=False, showgrid=False, showline=False),
        ),
    )
    return dict(data=data, layout=layout)


def write_feature_as_tsv(
    feature_data: np.ndarray, keys: typing.List[str], file_name: typing.Union[Path, str]
):
    with open(file_name, "w") as f:
        for i in range(feature_data.shape[0]):
            f.write(
                "\t".join([keys[i]] + [str(x) for x in list(feature_data[i])]) + "\n"
            )


def compress_object(raw_object, suite):
    return base64.b64encode(suite.encrypt(pickle.dumps(raw_object, protocol=4))).decode(
        "utf-8"
    )


def decompress_object(compressed_object, suite):
    return pickle.loads(suite.decrypt(base64.b64decode(compressed_object)))


def protein_to_aln_index(protein_index, aln_seq):
    n = 0
    for i in range(len(aln_seq)):
        if protein_index == n:
            return i
        elif aln_seq[i] == "-":
            pass
        else:
            n += 1


def aln_index_to_protein(alignment_index, alignment):
    res = dict()
    for k, v in alignment.items():
        if v[alignment_index] == "-":
            res[k] = None
        else:
            res[k] = alignment_index - v[:alignment_index].count("-")
    return res


def to_fasta_str(alignment):
    res = []
    for k, v in alignment.items():
        res.append(f">{k}")
        res.append(v)
    return "\n".join(res)


@dataclass
class PdbEntry:
    PDB_ID: str
    CHAIN_ID: str
    PdbResNumStart: int
    PdbResNumEnd: int
    PFAM_ACC: str
    PFAM_Name: str
    PFAM_desc: str
    eValue: float
    PDB_file: str = None

    @classmethod
    def from_data_line(cls, data_line):
        return cls(
            data_line[0],
            data_line[1],
            int(data_line[2]),
            int(data_line[3]),
            data_line[4],
            data_line[5],
            data_line[6],
            float(data_line[7]),
        )

    @classmethod
    def from_ali_header(cls, ali_header):
        headers = ali_header.split(":")
        if headers[3] == " ":
            chain_id = ""
        else:
            chain_id = headers[3]
        return cls(
            headers[1],
            chain_id,
            int(headers[2]),
            int(headers[4]),
            "none",
            "none",
            "none",
            0.0,
        )

    @classmethod
    def from_user_input(cls, pdb_path, chain_id="none"):
        return cls(
            Path(pdb_path).stem, chain_id, -1, -1, "none", "none", "none", 0.0, pdb_path
        )

    def get_pdb(self, from_atm_file=None):
        if from_atm_file is not None:
            pdb_obj = pd.parsePDB(
                f"{from_atm_file}{(self.PDB_ID + self.CHAIN_ID).lower()}.atm"
            )
        else:
            if self.PDB_file is not None:
                pdb_obj = pd.parsePDB(self.PDB_file)
            else:
                pdb_obj = pd.parsePDB(
                    self.PDB_ID, chain=self.CHAIN_ID
                )  # TODO : change mkdir and etc..
            if self.PdbResNumStart == -1 and self.PdbResNumEnd == -1:
                pdb_obj = pdb_obj.select(f"protein")
            else:
                pdb_obj = pdb_obj.select(
                    f"protein and resnum {self.PdbResNumStart} : {self.PdbResNumEnd + 1}"
                )
        if self.PDB_file is None:
            filename = f"{self.PDB_ID}_{self.CHAIN_ID}_{self.PdbResNumStart}.pdb"
            pd.writePDB(filename, pdb_obj)
        else:
            filename = self.PDB_file
        return pd.parsePDB(filename), filename

    def get_features(self, from_file=None):
        try:
            pdb_obj, filename = self.get_pdb(from_atm_file=from_file)
        except TypeError:
            return None, None
        return pdb_obj, filename

    @property
    def filename(self):
        if self.PDB_file is None:
            return f"{self.PDB_ID}_{self.CHAIN_ID}_{self.PdbResNumStart}.pdb"
        else:
            return self.PDB_file

    @property
    def unique_name(self):
        return f"{self.PDB_ID}_{self.CHAIN_ID}_{self.PdbResNumStart}"


class PfamToPDB:
    def __init__(
        self,
        uri="https://www.rcsb.org/pdb/rest/hmmer?file=hmmer_pdb_all.txt",
        from_file=False,
        limit=None,
    ):
        self.uri = uri
        if from_file:
            f = open("hmmer_pdb_all.txt", "r")
            raw_text = f.read()
            f.close()
        else:
            raw_text = rq.get(self.uri).text
        data_lines = [x.split("\t") for x in raw_text.split("\n")]
        self.headers = data_lines[0]
        data_lines = data_lines[1:]
        self.pfam_to_pdb_ids = dict()
        self._initiate_pfam_to_pdbids(data_lines, limit=limit)
        self.pdb_entries = None
        self.msa = None
        self.caretta_alignment = None

    def _initiate_pfam_to_pdbids(self, data_lines, limit=None):
        for line in data_lines:
            try:
                current_data = PdbEntry.from_data_line(line)
            except:
                continue
            if current_data.PFAM_ACC in self.pfam_to_pdb_ids:
                self.pfam_to_pdb_ids[current_data.PFAM_ACC].append(current_data)
            else:
                self.pfam_to_pdb_ids[current_data.PFAM_ACC] = [current_data]

        if limit:
            n = 0
            new = {}
            for k, v in self.pfam_to_pdb_ids.items():
                if n == limit:
                    break
                if len(v) < 4:
                    continue
                else:
                    new[k] = v
                    n += 1
            self.pfam_to_pdb_ids = new

    def get_entries_for_pfam(
        self, pfam_id, limit_by_score=1.0, limit_by_protein_number=50
    ):
        pdb_entries = list(
            filter(lambda x: (x.eValue < limit_by_score), self.pfam_to_pdb_ids[pfam_id])
        )[:limit_by_protein_number]
        return pdb_entries
